/*******************************************************************************

    File:     cdh_client.c
    Author:   Rick Cooper
    Created:  21/04/12
    Edited:   21/04/12
    Contents: LibExpress interface to the mirror model

    Public procedures:
        int main(int argc, char **argv)

*******************************************************************************/

#include "src_model.h"
#include "cdh_task.h"
#include "lib_maths.h"
#include <stdlib.h>
#include <time.h>
#include <gtk/gtk.h>
#include <libexpress.h>

#undef SIM1
#define SIM2

#define SERVER_PAGE "/rcooper/experiments/contingency_s2_rw/express.cgi"
#define LEARNING_RULE LR_RW_A

/******************************************************************************/

void update_graphics(GlobalVariables *gv) {}

/******************************************************************************/
/* Section 1: Main -----------------------------------------------------------*/

int main(int argc, char **argv)
{
    GlobalVariables *gv;
    int pid, trial, block;
#ifdef SIM1
    double fit_rms;
    double fit_max;
#endif
#ifdef SIM2
    double psp_score[3];
#endif

    if ((gv = globals_create()) == NULL) {
        fprintf(stderr, "%s: Aborting - Memory allocation failed while creating global variable space\n", argv[0]);
    }
    else {
#ifdef SIM1
        gv->subjects2 = 1; // Just one subject for simulation 1
        gv->subject_im_sd = 0.0; // No noise for best fit
#endif
#ifdef SIM2
        gv->subjects2 = 12; // 12 subjects for simulation 2
        gv->subject_im_sd = 1.0; // With noise so we have a distribution
#endif
        gv->context_input_strength[0] =  5.0;
        gv->context_input_strength[1] =  5.0;
        gv->context_input_strength[2] = 50.0;
        gv->context_input_strength[3] =  5.0;

        server_set(SP_HOST, "sideshow.psyc.bbk.ac.uk");
        server_set(SP_PAGE, SERVER_PAGE);
        gv->learning_rule = LEARNING_RULE;

        server_declare_variable("trial", VAR_WIV, VAR_INT, &trial);
        server_declare_variable("block", VAR_WIV, VAR_INT, &block);

        server_declare_variable("learning_rate", VAR_BIV, VAR_DOUBLE, &(gv->learning_rate));
        server_declare_variable("tau", VAR_BIV, VAR_DOUBLE, &(gv->target_value));

#ifdef SIM1
        server_declare_variable("cdh_fit_rms", VAR_DV, VAR_DOUBLE, &fit_rms);
        server_declare_variable("cdh_fit_max", VAR_DV, VAR_DOUBLE, &fit_max);
#endif
#ifdef SIM2
        server_declare_variable("cdh_psp1_score", VAR_DV, VAR_DOUBLE, &psp_score[0]);
        server_declare_variable("cdh_psp2_score", VAR_DV, VAR_DOUBLE, &psp_score[1]);
        server_declare_variable("cdh_psp3_score", VAR_DV, VAR_DOUBLE, &psp_score[2]);
#endif

        while ((pid = server_next_participant()) > 0) {
            do {
                exp2_do_task(gv);
#ifdef SIM1
                fit_rms = exp2_fit_rms(gv);
                fit_max = exp2_fit_max(gv);
#endif
#ifdef SIM2
                psp_score[0] = exp2_fit_psp_score(gv, 0);
                psp_score[1] = exp2_fit_psp_score(gv, 1);
                psp_score[2] = exp2_fit_psp_score(gv, 2);
#endif
            } while (server_next_hit(pid) > 0);
        }
    }
    globals_destroy(gv);
    server_free();

    exit(0);
}

/******************************************************************************/
