#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gtk/gtk.h>
#include <cairo/cairo-pdf.h>
#include "lib/lib_gtkx.h"
#include "lib/lib_cairox.h"

/******************************************************************************/

typedef enum dv {
    DV_CH1_FIT_RMS, DV_CH1_FIT_RSQ, DV_CH2_FIT_RMS, DV_CH2_FIT_RSQ, DV_MAX
} DV;

#define DELAY_S_MAX 21
#define DELAY_I_MAX 21
#define STRENGTH_S_MAX 21
#define STRENGTH_I_MAX 21

#define W_PIXELS (1 + (DELAY_S_MAX - 1) * 10)
#define H_PIXELS (1 + (DELAY_I_MAX - 1) * 10)

#define PID_MAX 101

//#define DATA_FILE "sim1_delay.dat"
//#define DATA_FILE "sim1_strength.dat"

#undef COLOUR

int counts[DELAY_S_MAX][DELAY_I_MAX];
double simulated_values[PID_MAX][DELAY_S_MAX][DELAY_I_MAX][DV_MAX];
double simulated_mean[DELAY_S_MAX][DELAY_I_MAX][DV_MAX];

int delay_s_values[DELAY_S_MAX];
int delay_i_values[DELAY_I_MAX];
double strength_s_values[STRENGTH_S_MAX];
double strength_i_values[STRENGTH_I_MAX];
int s_index[DELAY_S_MAX];
int i_index[DELAY_I_MAX];

double data[W_PIXELS][H_PIXELS];

/******************************************************************************/


/* We use this to indicate that an IV's value has not been set */
#define MAX_NEG -999999

/*----------------------------------------------------------------------------*/

static double dabs(double x)
{
    return((x > 0) ? x : -x);
}

static int map_strength_s(double strength_s)
{
    int i;

    for (i = 0; i < STRENGTH_S_MAX; i++) {
        if (strength_s_values[i] == MAX_NEG) {
	    strength_s_values[i] = strength_s;
	    return(i);
	}
	else if (dabs(strength_s_values[i] - strength_s) < 0.00001) {
	    return(i);
	}
    }
    return(-1);
}

static int map_strength_i(double strength_i)
{
    int i;

    for (i = 0; i < STRENGTH_I_MAX; i++) {
        if (strength_i_values[i] == MAX_NEG) {
	    strength_i_values[i] = strength_i;
	    return(i);
	}
	else if (dabs(strength_i_values[i] - strength_i) < 0.00001) {
	    return(i);
	}
    }
    return(-1);
}

static int map_delay_s(int delay_s)
{
    int i;

    for (i = 0; i < DELAY_S_MAX; i++) {
        if (delay_s_values[i] == MAX_NEG) {
	    delay_s_values[i] = delay_s;
	    return(i);
	}
	else if (dabs(delay_s_values[i] - delay_s) < 0.00001) {
	    return(i);
	}
    }
    return(-1);
}

static int map_delay_i(int delay_i)
{
    int i;

    for (i = 0; i < DELAY_I_MAX; i++) {
        if (delay_i_values[i] == MAX_NEG) {
	    delay_i_values[i] = delay_i;
	    return(i);
	}
	else if (dabs(delay_i_values[i] - delay_i) < 0.00001) {
	    return(i);
	}
    }
    return(-1);
}

/*----------------------------------------------------------------------------*/

static void initialise_counts()
{
    int i, j;

    for (i = 0; i < DELAY_S_MAX; i++) {
        for (j = 0; j < DELAY_I_MAX; j++) {
                counts[i][j] = 0;
        }
    }
    for (i = 0; i < DELAY_S_MAX; i++) {
        delay_s_values[i] = MAX_NEG;
    }
    for (j = 0; j < DELAY_I_MAX; j++) {
        delay_i_values[j] = MAX_NEG;
    }
    for (i = 0; i < STRENGTH_S_MAX; i++) {
        strength_s_values[i] = MAX_NEG;
    }
    for (j = 0; j < STRENGTH_I_MAX; j++) {
        strength_i_values[j] = MAX_NEG;
    }
}

static void ignore_to_eoln(FILE *fp)
{
    int c;

    do {
        c = getc(fp);
    } while ((c > 0) && (c != '\n'));
}

static int read_data_line_delay(FILE *fp)
{
    double ch1_fit_rms, ch1_fit_rsq, ch2_fit_rms, ch2_fit_rsq;
    double strength_i, strength_s;
    int pid, delay_s, delay_i;

    if (fscanf(fp, "%d %d %d %lf %lf %lf %lf %lf %lf\n", &pid, &delay_i, &delay_s, &strength_i, &strength_s, &ch1_fit_rms, &ch1_fit_rsq, &ch2_fit_rms, &ch2_fit_rsq) > 0) {
        int s = map_delay_s(delay_s);
        int i = map_delay_i(delay_i);

        if (s < 0) {
            fprintf(stdout, "Too many spatial delays\n");
        }
        else if (i < 0) {
            fprintf(stdout, "Too many imitative delays\n");
        }
        else if (counts[s][i] < PID_MAX) {
            simulated_values[counts[s][i]][s][i][DV_CH1_FIT_RMS] = ch1_fit_rms;
            simulated_values[counts[s][i]][s][i][DV_CH1_FIT_RSQ] = ch1_fit_rsq;
            simulated_values[counts[s][i]][s][i][DV_CH2_FIT_RMS] = ch2_fit_rms;
            simulated_values[counts[s][i]][s][i][DV_CH2_FIT_RSQ] = ch2_fit_rsq;
            counts[s][i]++;
        }
        else {
            fprintf(stdout, "Too many participants (delay_i: %d; delay_s: %d)\n", delay_i, delay_s);
        }
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

static int read_data_line_strength(FILE *fp)
{
    double ch1_fit_rms, ch1_fit_rsq, ch2_fit_rms, ch2_fit_rsq;
    double strength_i, strength_s;
    int pid, delay_s, delay_i;

    if (fscanf(fp, "%d %d %d %lf %lf %lf %lf %lf %lf\n", &pid, &delay_i, &delay_s, &strength_i, &strength_s, &ch1_fit_rms, &ch1_fit_rsq, &ch2_fit_rms, &ch2_fit_rsq) > 0) {
        int s = map_strength_s(strength_s);
        int i = map_strength_i(strength_i);

        if (s < 0) {
            fprintf(stdout, "Too many spatial delays\n");
        }
        else if (i < 0) {
            fprintf(stdout, "Too many imitative delays\n");
        }
        else if (counts[s][i] < PID_MAX) {
// fprintf(stdout, "%d %d %f %f %f %f\n", s, i, ch1_fit_rms, ch1_fit_rsq, ch2_fit_rms, ch2_fit_rsq);
            simulated_values[counts[s][i]][s][i][DV_CH1_FIT_RMS] = ch1_fit_rms;
            simulated_values[counts[s][i]][s][i][DV_CH1_FIT_RSQ] = ch1_fit_rsq;
            simulated_values[counts[s][i]][s][i][DV_CH2_FIT_RMS] = ch2_fit_rms;
            simulated_values[counts[s][i]][s][i][DV_CH2_FIT_RSQ] = ch2_fit_rsq;
            counts[s][i]++;
        }
        else {
            fprintf(stdout, "Too many participants (strength_i: %f; strength_s: %f)\n", strength_i, strength_s);
        }
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

int data_read_delay(char *filename)
{
    int count;
    FILE *fp;

    initialise_counts();
    if ((fp = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "Failed to open %s\n", filename);
        return(0);
    }
    else {
        ignore_to_eoln(fp);
        count = 0;
        while (read_data_line_delay(fp)) {
            count++;
        }
        fprintf(stdout, "Read %d records from %s\n", count, filename);
        return(1);
    }
}

int data_read_strength(char *filename)
{
    int count;
    FILE *fp;

    initialise_counts();
    if ((fp = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "Failed to open %s\n", filename);
        return(0);
    }
    else {
        ignore_to_eoln(fp);
        count = 0;
        while (read_data_line_strength(fp)) {
            count++;
        }
        fprintf(stdout, "Read %d records from %s\n", count, filename);
        return(1);
    }
}

/******************************************************************************/

static double scale_rms(double data)
{
#ifdef COLOUR
    if (data > 15) {
        return(1.00);
    }
    else if (data < 5) {
        return(0.00);
    }
    else {
        return((data - 5) / 10.0);
    }
#else
    if (data < 6.0) {
        return(0.999);
    }
    else if (data < 7.0) {
        return(0.888);
    }
    else if (data < 8.0) {
        return(0.777);
    }
    else if (data < 9.0) {
        return(0.666);
    }
    else if (data < 10.0) {
        return(0.555);
    }
    else if (data < 11.0) {
        return(0.444);
    }
    else if (data < 12.0) {
        return(0.333);
    }
    else if (data < 13.0) {
        return(0.222);
    }
    else if (data < 14.0) {
        return(0.111);
    }
    else {
        return(0.0000);
    }
#endif
}

static double scale_rsq(double data)
{
#ifdef COLOUR
    if (data > 1.0) {
        return(1.0);
    }
    else if (data < 0.5) {
        return(0.0);
    }
    else {
        return((data - 0.5) * 2.0);
    }
#else
    if (data > 0.95) {
        return(0.999);
    }
    else if (data > 0.90) {
        return(0.888);
    }
    else if (data > 0.85) {
        return(0.777);
    }
    else if (data > 0.80) {
        return(0.666);
    }
    else if (data > 0.75) {
        return(0.555);
    }
    else if (data > 0.70) {
        return(0.444);
    }
    else if (data > 0.65) {
        return(0.333);
    }
    else if (data > 0.60) {
        return(0.222);
    }
    else if (data > 0.55) {
        return(0.111);
    }
    else {
        return(0.000);
    }
#endif
}

/******************************************************************************/

static void collate_simulated_dvs_delay()
{
    int i, j, m, n;
    
    for (i = 0; i < DELAY_S_MAX; i++) {
        for (j = 0; j < DELAY_I_MAX; j++) {
            for (m = 0; m < DV_MAX; m++) {
                if (counts[i][j] > 0) {
                    double sum = 0;

                    for (n = 0; n < counts[i][j]; n++) {
                        sum += simulated_values[n][i][j][m];
                    }
                    simulated_mean[i][j][m] = sum / (double) counts[i][j];
                }
                else {
                    simulated_mean[i][j][m] = 0.0;
                }
            }
        }
    }
}

static void collate_simulated_dvs_strength()
{
    int i, j, m, n;
    
    for (i = 0; i < STRENGTH_S_MAX; i++) {
        for (j = 0; j < STRENGTH_I_MAX; j++) {
            for (m = 0; m < DV_MAX; m++) {
                if (counts[i][j] > 0) {
                    double sum = 0;

                    for (n = 0; n < counts[i][j]; n++) {
                        sum += simulated_values[n][i][j][m];
                    }
                    simulated_mean[i][j][m] = sum / (double) counts[i][j];
                }
                else {
                    simulated_mean[i][j][m] = 0.0;
                }
            }
        }
    }
}

static void generate_indicies_delay()
{
    // Populate arrays with rank-ordered indices for the IVs:

    int i, j, k;

    for (i = 0; i < DELAY_S_MAX; i++) {
        s_index[i] = i;
    }
    for (i = 0; i < DELAY_S_MAX; i++) {
        for (j = i+1; j < DELAY_S_MAX; j++) {
            if (delay_s_values[s_index[i]] > delay_s_values[s_index[j]]) {
                k = s_index[i]; s_index[i] = s_index[j]; s_index[j] = k;
	    }
        }
    }

    for (i = 0; i < DELAY_I_MAX; i++) {
        i_index[i] = i;
    }
    for (i = 0; i < DELAY_I_MAX; i++) {
        for (j = i+1; j < DELAY_I_MAX; j++) {
            if (delay_i_values[i_index[i]] > delay_i_values[i_index[j]]) {
                k = i_index[i]; i_index[i] = i_index[j]; i_index[j] = k;
	    }
        }
    }
}


static void generate_indicies_strength()
{
    // Populate arrays with rank-ordered indices for the IVs:

    int i, j, k;

    for (i = 0; i < STRENGTH_S_MAX; i++) {
        s_index[i] = i;
    }
    for (i = 0; i < STRENGTH_S_MAX; i++) {
        for (j = i+1; j < STRENGTH_S_MAX; j++) {
            if (strength_s_values[s_index[i]] > strength_s_values[s_index[j]]) {
                k = s_index[i]; s_index[i] = s_index[j]; s_index[j] = k;
	    }
        }
    }

    for (i = 0; i < STRENGTH_I_MAX; i++) {
        i_index[i] = i;
    }
    for (i = 0; i < STRENGTH_I_MAX; i++) {
        for (j = i+1; j < STRENGTH_I_MAX; j++) {
            if (strength_i_values[i_index[i]] > strength_i_values[i_index[j]]) {
                k = i_index[i]; i_index[i] = i_index[j]; i_index[j] = k;
	    }
        }
    }
}

static void draw_grid_delay(cairo_t *cr, PangoLayout *layout, char *graph_type, int x0, int y0)
{
    CairoxTextParameters p;
    char buffer[32];
    int i;

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    /* Vertical grid */
    for (i = 0; i < 11; i++) {
        cairox_paint_line(cr, 0.5, x0 + i * 10, y0, x0 + i * 10, y0 + 100);
    }

    /* Horizontal grid */
    for (i = 0; i < 11; i++) {
        cairox_paint_line(cr, 0.5, x0, y0 + i * 10, x0 + 100, y0 + i * 10);
    }

    /* X (horizontal) axis labels: */

    pangox_layout_set_font_size(layout, 7 / 1.5);
    for (i = 0; i < 11; i++) {
        sprintf(buffer, "%3d", (int) (delay_s_values[s_index[0]] + i * (delay_s_values[s_index[DELAY_S_MAX-1]] - delay_s_values[s_index[0]]) / 10.0));
        cairox_text_parameters_set(&p, x0+i*10, y0+103, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    pangox_layout_set_font_size(layout, 12 / 1.5);
    cairox_text_parameters_set(&p, x0+50, y0+112, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    sprintf(buffer, "%s (spatial)", graph_type);
    cairox_paint_pango_text(cr, &p, layout, buffer);

    /* Y (vertical) axis labels: */

    pangox_layout_set_font_size(layout, 7 / 1.5);
    for (i = 0; i < 11; i++) {
        sprintf(buffer, "%3d", (int) (delay_i_values[i_index[0]] + i * (delay_i_values[i_index[DELAY_I_MAX-1]] - delay_i_values[i_index[0]]) / 10.0));
        cairox_text_parameters_set(&p, x0-3, y0+100-i*10, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    pangox_layout_set_font_size(layout, 12 / 1.5);
    cairox_text_parameters_set(&p, x0-23, y0+50, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 90.0);
    sprintf(buffer, "%s (imitative)", graph_type);
    cairox_paint_pango_text(cr, &p, layout, buffer);
}

static void draw_grid_strength(cairo_t *cr, PangoLayout *layout, char *graph_type, int x0, int y0)
{
    CairoxTextParameters p;
    char buffer[32];
    int i;

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    /* Vertical grid */
    for (i = 0; i < 11; i++) {
        cairox_paint_line(cr, 0.5, x0 + i * 10, y0, x0 + i * 10, y0 + 100);
    }

    /* Horizontal grid */
    for (i = 0; i < 11; i++) {
        cairox_paint_line(cr, 0.5, x0, y0 + i * 10, x0 + 100, y0 + i * 10);
    }

    /* X (horizontal) axis labels: */

    pangox_layout_set_font_size(layout, 7 / 1.5);
    for (i = 0; i < 11; i++) {
        sprintf(buffer, "%3.1f", (strength_s_values[s_index[0]] + i * (strength_s_values[s_index[DELAY_S_MAX-1]] - strength_s_values[s_index[0]]) / 10.0));
        cairox_text_parameters_set(&p, x0+i*10, y0+103, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    pangox_layout_set_font_size(layout, 12 / 1.5);
    cairox_text_parameters_set(&p, x0+50, y0+112, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    sprintf(buffer, "%s (spatial)", graph_type);
    cairox_paint_pango_text(cr, &p, layout, buffer);

    /* Y (vertical) axis labels: */

    pangox_layout_set_font_size(layout, 7 / 1.5);
    for (i = 0; i < 11; i++) {
        sprintf(buffer, "%3.1f", (strength_i_values[i_index[0]] + i * (strength_i_values[i_index[DELAY_I_MAX-1]] - strength_i_values[i_index[0]]) / 10.0));
        cairox_text_parameters_set(&p, x0-3, y0+100-i*10, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    pangox_layout_set_font_size(layout, 12 / 1.5);
    cairox_text_parameters_set(&p, x0-23, y0+50, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 90.0);
    sprintf(buffer, "%s (imitative)", graph_type);
    cairox_paint_pango_text(cr, &p, layout, buffer);
}

static void draw_title(cairo_t *cr, PangoLayout *layout, double x0, double y0, char *title)
{
    CairoxTextParameters p;

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_rectangle(cr, x0, y0, 100, 100);
    cairo_stroke(cr);

    pangox_layout_set_font_size(layout, 14 / 1.5);
    cairox_text_parameters_set(&p, x0+50, y0-4, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, title);
}

static void draw_title_superscript(cairo_t *cr, PangoLayout *layout, double x0, double y0, char *ss)
{
    CairoxTextParameters p;

    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    pangox_layout_set_font_size(layout, 11 / 1.5);
    cairox_text_parameters_set(&p, x0+6, y0-8, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, ss);
}

static void draw_legend_rms(cairo_t *cr, PangoLayout *layout, int x0, int y0)
{
    CairoxTextParameters p;
    char buffer[8];
    int i;

#ifdef COLOUR
    for (i = 0; i < 100; i++) {
        double r, g, b;

        get_rgb_colour(i / 100.0, &r, &g, &b);
        cairo_set_source_rgb(cr, r, g, b);
        cairo_rectangle(cr, x0-3, y0+i-0.1, 6, 1.2);
        cairo_fill(cr);
    }
#else
    for (i = 0; i < 10; i++) {
        cairo_set_source_rgb(cr, (i/9.0), (i/9.0), (i/9.0));
        cairo_rectangle(cr, x0-3, y0+i*10, 6, 10);
        cairo_fill(cr);
    }
#endif

    pangox_layout_set_font_size(layout, 8 / 1.5);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_rectangle(cr, x0-3, y0, 6, 100);
    cairo_stroke(cr);

    for (i = 0; i < 6; i++) {
        cairox_paint_line(cr, 1.0, x0-2.9, y0+i*20.0, x0+5.0, y0+i*20.0);
        sprintf(buffer, "%2d", 15 - (i * 2));
        cairox_text_parameters_set(&p, x0+6, y0+i*20, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }
}

static void draw_legend_rsq(cairo_t *cr, PangoLayout *layout, int x0, int y0)
{
    CairoxTextParameters p;
    char buffer[8];
    int i;

#ifdef COLOUR
    for (i = 0; i < 100; i++) {
        double r, g, b;

        get_rgb_colour(i / 100.0, &r, &g, &b);
        cairo_set_source_rgb(cr, r, g, b);
        cairo_rectangle(cr, x0-3, y0+i-0.1, 6, 1.2);
        cairo_fill(cr);
    }
#else
    for (i = 0; i < 10; i++) {
        cairo_set_source_rgb(cr, 1.0-(i/9.0), 1.0-(i/9.0), 1.0-(i/9.0));
        cairo_rectangle(cr, x0-3, y0+i*10, 6, 10);
        cairo_fill(cr);
    }
#endif

    pangox_layout_set_font_size(layout, 8 / 1.5);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_rectangle(cr, x0-3, y0, 6, 100);
    cairo_stroke(cr);

    for (i = 0; i < 6; i++) {
        cairox_paint_line(cr, 1.0, x0-2.9, y0+i*20.0, x0+5.0, y0+i*20.0);
        sprintf(buffer, "%4.1f", 1.0 - i / 10.0);
        cairox_text_parameters_set(&p, x0+6, y0+i*20, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }
}

/******************************************************************************/

static void load_graph_data(int dv)
{
    int COMPRESSION_X = (W_PIXELS-1) / (DELAY_S_MAX-1);
    int COMPRESSION_Y = (H_PIXELS-1) / (DELAY_I_MAX-1);
    int i, j;

    for (i = 0; i < W_PIXELS; i++) {
        for (j = 0; j < H_PIXELS; j++) {
            int xw1, xw2, yw1, yw2;
            int ei1 = 0, ei2 = 0, ii1 = 0, ii2 = 0;

            /* Weights: */
            xw1 = COMPRESSION_X * (1 + (i / COMPRESSION_X)) - i;
            xw2 = i - COMPRESSION_X * (i / COMPRESSION_X);

            yw1 = COMPRESSION_Y * (1 + (j / COMPRESSION_Y)) - j;
            yw2 = j - COMPRESSION_Y * (j / COMPRESSION_Y);

            /* Indices: */
            if (xw1 != 0) {
                ei1 = s_index[i / COMPRESSION_X];
            }
            if (xw2 != 0) {
                ei2 = s_index[1 + (i / COMPRESSION_X)];
            }
            if (yw1 != 0) {
                ii1 = i_index[j / COMPRESSION_Y];
            }
            if (yw2 != 0) {
                ii2 = i_index[1 + (j / COMPRESSION_Y)];
            }

            data[i][j] = (xw1 * yw1 * simulated_mean[ei1][ii1][dv] + 
                          xw1 * yw2 * simulated_mean[ei1][ii2][dv] + 
                          xw2 * yw1 * simulated_mean[ei2][ii1][dv] + 
                          xw2 * yw2 * simulated_mean[ei2][ii2][dv]) / 
                        ((xw1 + xw2) * (yw1 + yw2));

        }
    }
}

static void draw_graph(cairo_t *cr, int x0, int y0, double scale())
{
    double voxel_W = 101 / (double) W_PIXELS;
    double voxel_H = 101 / (double) H_PIXELS;
    int i, j;

    for (i = 0; i < W_PIXELS; i++) {
        for (j = 0; j < H_PIXELS; j++) {
#ifdef COLOUR
            double r, g, b;

            get_rgb_colour(scale(data[i][j]), &r, &g, &b);
            cairo_set_source_rgb(cr, r, g, b);
#else
            double d = scale(data[i][j]);
            cairo_set_source_rgb(cr, d, d, d);
#endif
            cairo_rectangle(cr, x0 + i * voxel_W, y0 + 100 - j * voxel_H, 1.0, 1.0);
            cairo_fill(cr);
	}
    }
}

/******************************************************************************/

static void draw_delay_surface_ch1_rms(cairo_t *cr, PangoLayout *layout, double x0, double y0)
{
    load_graph_data(DV_CH1_FIT_RMS);
    draw_graph(cr, x0, y0, scale_rms);
    draw_grid_delay(cr, layout, "Delay", x0, y0);
    draw_title(cr, layout, x0, y0, "RMS fit to behavioral data");
    draw_legend_rms(cr, layout, x0+110, y0);
}

static void draw_delay_surface_ch1_rsq(cairo_t *cr, PangoLayout *layout, double x0, double y0)
{
    load_graph_data(DV_CH1_FIT_RSQ);
    draw_graph(cr, x0, y0, scale_rsq);
    draw_grid_delay(cr, layout, "Delay", x0, y0);
    draw_title(cr, layout, x0, y0, "R  fit to behavioral data");
    draw_title_superscript(cr, layout, x0, y0, "2");
    draw_legend_rsq(cr, layout, x0+110, y0);
}

static void draw_delay_surface_ch2_rms(cairo_t *cr, PangoLayout *layout, double x0, double y0)
{
    load_graph_data(DV_CH2_FIT_RMS);
    draw_graph(cr, x0, y0, scale_rms);
    draw_grid_delay(cr, layout, "Delay", x0, y0);
    draw_title(cr, layout, x0, y0, "RMS fit to behavioral data");
    draw_legend_rms(cr, layout, x0+110, y0);
}

static void draw_delay_surface_ch2_rsq(cairo_t *cr, PangoLayout *layout, double x0, double y0)
{
    load_graph_data(DV_CH2_FIT_RSQ);
    draw_graph(cr, x0, y0, scale_rsq);
    draw_grid_delay(cr, layout, "Delay", x0, y0);
    draw_title(cr, layout, x0, y0, "R  fit to behavioral data");
    draw_title_superscript(cr, layout, x0, y0, "2");
    draw_legend_rsq(cr, layout, x0+110, y0);
}

static void draw_strength_surface_ch1_rms(cairo_t *cr, PangoLayout *layout, double x0, double y0)
{
    load_graph_data(DV_CH1_FIT_RMS);
    draw_graph(cr, x0, y0, scale_rms);
    draw_grid_strength(cr, layout, "Strength", x0, y0);
    draw_title(cr, layout, x0, y0, "RMS fit to behavioral data");
    draw_legend_rms(cr, layout, x0+110, y0);
}

static void draw_strength_surface_ch1_rsq(cairo_t *cr, PangoLayout *layout, double x0, double y0)
{
    load_graph_data(DV_CH1_FIT_RSQ);
    draw_graph(cr, x0, y0, scale_rsq);
    draw_grid_strength(cr, layout, "Strength", x0, y0);
    draw_title(cr, layout, x0, y0, "R  fit to behavioral data");
    draw_title_superscript(cr, layout, x0, y0, "2");
    draw_legend_rsq(cr, layout, x0+110, y0);
}

static void draw_strength_surface_ch2_rms(cairo_t *cr, PangoLayout *layout, double x0, double y0)
{
    load_graph_data(DV_CH2_FIT_RMS);
    draw_graph(cr, x0, y0, scale_rms);
    draw_grid_strength(cr, layout, "Strength", x0, y0);
    draw_title(cr, layout, x0, y0, "RMS fit to behavioral data");
    draw_legend_rms(cr, layout, x0+110, y0);
}

static void draw_strength_surface_ch2_rsq(cairo_t *cr, PangoLayout *layout, double x0, double y0)
{
    load_graph_data(DV_CH2_FIT_RSQ);
    draw_graph(cr, x0, y0, scale_rsq);
    draw_grid_strength(cr, layout, "Strength", x0, y0);
    draw_title(cr, layout, x0, y0, "R  fit to behavioral data");
    draw_title_superscript(cr, layout, x0, y0, "2");
    draw_legend_rsq(cr, layout, x0+110, y0);
}

/******************************************************************************/

static void maps_draw(GlobalVariables *gv, cairo_t *cr, int width, int height)
{
    PangoLayout *layout;

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    cairo_save(cr);
    cairo_scale(cr, 1.5, 1.5);

    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    if (data_read_delay("../Results/sim1_delay.dat")) {
        collate_simulated_dvs_delay();
        generate_indicies_delay();
        draw_delay_surface_ch2_rms(cr, layout, 40, 20);
        draw_delay_surface_ch2_rsq(cr, layout, 220, 20);
    }

    if (data_read_strength("../Results/sim1_strength.dat")) {
        collate_simulated_dvs_strength();
        generate_indicies_strength();
        draw_strength_surface_ch2_rms(cr, layout, 40, 180);
        draw_strength_surface_ch2_rsq(cr, layout, 220, 180);
    }

    g_object_unref(layout);
    cairo_restore(cr);
}

/******************************************************************************/

void maps_create(GlobalVariables *globals, int width, int height)
{
    cairo_surface_t *surface;
    cairo_t *cr;

    surface = cairo_pdf_surface_create("../Figures/image_maps.pdf", width, height);
    cr = cairo_create(surface);
    maps_draw(gv, cr, width, height);
    cairo_destroy(cr);
    cairo_surface_destroy(surface);
}

/******************************************************************************/
