#include "src_model.h"
#include "cdh_task.h"
#include "lib_maths.h"
#include "lib_gtkx.h"
#include "lib_cairox.h"

#define TRAINING

#undef DRAW_ASSOCIATION_TABLE
#undef COLOUR
#define COL_WIDTH 83

/******************************************************************************/

GtkWidget *gtkx_space(GtkWidget *box)
{
    GtkWidget *tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(box), tmp, TRUE, TRUE, 0);
    gtk_widget_show(tmp);
    return(tmp);
}

/******************************************************************************/
/* SECTION 1: The network architecture ****************************************/
/******************************************************************************/

#define YMARGIN 80

void cairox_select_grey(cairo_t *cr, int i, double alpha)
{
    if (i < 0) {        /* White */
        cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, alpha);
    }
    else if (i == 0) {  /* Black */
        cairo_set_source_rgba(cr, 0.0, 0.0, 0.0, alpha);
    }
    else if (i == 1) {  /* Gray */
        cairo_set_source_rgba(cr, 0.7, 0.7, 0.7, alpha);
    }
    else if (i == 2) {  /* Gray 2 */
        cairo_set_source_rgba(cr, 0.5, 0.5, 0.5, alpha);
    }
    else if (i == 3) {  /* Gray 3 */
        cairo_set_source_rgba(cr, 0.2, 0.2, 0.2, alpha);
    }
    else if (i == 4) {  /* Gray 4 */
        cairo_set_source_rgba(cr, 0.8, 0.8, 0.8, alpha);
    }
}

static void draw_node(cairo_t *cr, PangoLayout *layout, double x, double y, GlobalVariables *gv, int i, Node *node)
{
    CairoxTextParameters p;
    char buffer[8];

    cairo_arc(cr, x, y, 20, 0, 2 * M_PI);
    cairo_set_source_rgb(cr, 1, 1, 1);
    cairo_fill_preserve(cr);
#ifdef COLOUR
    cairox_select_colour(cr, node->colour, node->activation[gv->cycle]);
#else
    cairo_set_source_rgba(cr, 0, 0, 0, node->activation[gv->cycle]);
#endif
    cairo_fill_preserve(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_set_line_width(cr, 1.0);
    cairo_stroke(cr);

    g_snprintf(buffer, 8, "%4.2f", node->activation[gv->cycle]);
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x, y, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, buffer);
}

static double line_length(CairoxPoint *p1, CairoxPoint *p2)
{
    return(sqrt((p1->x-p2->x)*(p1->x-p2->x) + (p1->y-p2->y)*(p1->y-p2->y)));
}

static void draw_li(cairo_t *cr, double li, double x1, double y1, double x2, double y2)
{
    CairoxPoint pc, p1, p2;

    if (li > 0.0001) {
#ifdef COLOUR
        cairo_set_source_rgb(cr, 0.0, 0.0, 1.0);
#else
        cairo_set_source_rgb(cr, 0.2, 0.2, 0.2);
#endif
        if (y1 < 200) {
            /* Put arc below nodes */
            p1.x = x1+24*cos(M_PI/4.0);
            p1.y = y1+24*sin(M_PI/4.0);
            p2.x = x2-24*cos(M_PI/4.0);
            p2.y = y2+24*sin(M_PI/4.0);
            pc.x = (x1+x2)/2.0;
            pc.y = (y1+y2)/2.0-(x2-x1)/3.0;

            cairo_set_line_width(cr, 1.0);
            cairo_arc(cr, p1.x, p1.y, 4.0, 0, 2 * M_PI);
            cairo_fill(cr);
            cairo_arc(cr, p2.x, p2.y, 4.0, 0, 2 * M_PI);
            cairo_fill(cr);

            cairo_set_line_width(cr, li);
            cairo_arc_negative(cr, pc.x, pc.y, line_length(&pc, &p1), line_calculate_slope(&p1, &pc), line_calculate_slope(&p2, &pc));
            cairo_stroke(cr);
        }
        else {
            /* Put arc above nodes */
            p1.x = x1+24*cos(M_PI/4.0);
            p1.y = y1-24*sin(M_PI/4.0);
            p2.x = x2-24*cos(M_PI/4.0);
            p2.y = y2-24*sin(M_PI/4.0);
            pc.x = (x1+x2)/2.0;
            pc.y = (y1+y2)/2.0+(x2-x1)/3.0;

            cairo_set_line_width(cr, 1.0);
            cairo_arc(cr, p1.x, p1.y, 4.0, 0, 2 * M_PI);
            cairo_fill(cr);
            cairo_arc(cr, p2.x, p2.y, 4.0, 0, 2 * M_PI);
            cairo_fill(cr);

            cairo_set_line_width(cr, li);
            cairo_arc(cr, pc.x, pc.y, line_length(&p2, &pc), line_calculate_slope(&p1, &pc), line_calculate_slope(&p2, &pc));
            cairo_stroke(cr);
        }
    }
}

static void cairo_draw_ellipse(cairo_t *cr, double x, double y, double aspect)
{
    cairo_save(cr);
    cairo_translate(cr, x, y);
    cairo_scale(cr, 60*aspect, 60);
    cairo_arc(cr, 0, 0, 1, 0, 2*M_PI);
    cairo_restore(cr);
    cairo_stroke(cr);
}

static void cairo_draw_network(cairo_t *cr, GlobalVariables *gv)
{
    CairoxTextParameters p;
    PangoLayout *layout;
    int width, height;
    char buffer[64];
    double x[2], dx[2], y[2];
    double xi, xj;
    int i, j;

    width = GTK_WIDGET(gv->canvas_network)->allocation.width;
    height = GTK_WIDGET(gv->canvas_network)->allocation.height;

    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    x[0] = width / (2.0 * (NUM_SENSE+NUM_CONTEXT+NUM_IMPERATIVE));
    x[1] = 80 + (width - 160) / (2.0 * NUM_MOTOR);
    dx[0] = width / (1.0 * (NUM_SENSE+NUM_CONTEXT+NUM_IMPERATIVE));
    dx[1] = (width - 160) / (1.0 * NUM_MOTOR);

    y[0] = 80 + YMARGIN;
    y[1] = height - 60 - YMARGIN;

    /* Ellipses around each set of nodes: */
    cairo_set_source_rgb(cr, 0.5, 0.5, 0.5);
    /* Sensory nodes */
    cairo_draw_ellipse(cr, x[0]+(NUM_SENSE-1)/2.0*dx[0], y[0], width * 0.0019);
    /* Context nodes */
    cairo_draw_ellipse(cr, x[0]+(NUM_SENSE+(NUM_CONTEXT-1)/2.0)*dx[0], y[0], width * 0.0020*(NUM_CONTEXT/2.0));
    /* Imperative nodes */
    cairo_draw_ellipse(cr, x[0]+(NUM_SENSE+NUM_CONTEXT+(NUM_IMPERATIVE-1)/2.0)*dx[0], y[0], width * 0.0019);
    /* Response nodes: */
    cairo_draw_ellipse(cr, x[1]+0.5*dx[1], y[1], width * 0.0049);

    /* The associative links: */

    for (j = 0; j < NUM_MOTOR; j++) {
        double weight;

        xj = x[1] + j * dx[1];
        for (i = 0; i < NUM_SENSE; i++) {
            xi = x[0] + i * dx[0];
            weight = WEIGHT_SM(j,i);
            if (weight > 0.0) {
#ifdef COLOUR
                cairo_set_source_rgb(cr, 1.0, 0.0, 0.0);
#else
                cairo_set_source_rgb(cr, 0.2, 0.2, 0.2);
#endif
                cairox_paint_line(cr, weight, xj, y[1], xi, y[0]);
            }
            else if (weight < 0.0) {
#ifdef COLOUR
                cairo_set_source_rgb(cr, 0.0, 0.0, 1.0);
#else
                cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
#endif
                cairox_paint_line(cr, -weight, xj, y[1], xi, y[0]);
            }
        }
        for (i = 0; i < NUM_CONTEXT; i++) {
            xi = x[0] + (NUM_SENSE+i) * dx[0];
            weight = WEIGHT_CM(j,i);
            if (weight > 0.0) {
#ifdef COLOUR
                cairo_set_source_rgb(cr, 1.0, 0.0, 0.0);
#else
                cairo_set_source_rgb(cr, 0.2, 0.2, 0.2);
#endif
                cairox_paint_line(cr, weight, xj, y[1], xi, y[0]);
            }
            else if (weight < 0.0) {
#ifdef COLOUR
                cairo_set_source_rgb(cr, 0.0, 0.0, 1.0);
#else
                cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
#endif
                cairox_paint_line(cr, -weight, xj, y[1], xi, y[0]);
            }
        }
        for (i = 0; i < NUM_IMPERATIVE; i++) {
            xi = x[0] + (NUM_SENSE+NUM_CONTEXT+i) * dx[0];
            weight = WEIGHT_IM(j,i);
            if (weight > 0.0) {
#ifdef COLOUR
                cairo_set_source_rgb(cr, 1.0, 0.0, 0.0);
#else
                cairo_set_source_rgb(cr, 0.2, 0.2, 0.2);
#endif
                cairox_paint_line(cr, weight, xj, y[1], xi, y[0]);
            }
            else if (weight < 0.0) {
#ifdef COLOUR
                cairo_set_source_rgb(cr, 0.0, 0.0, 1.0);
#else
                cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
#endif
                cairox_paint_line(cr, -weight, xj, y[1], xi, y[0]);
            }
        }
    }

    /* The lateral inhibition links: */

    xi = x[0]; xj = xi + dx[0];
    draw_li(cr, gv->sensory_li,  xi, y[0], xj, y[0]);

    xi = x[0] + 2*dx[0]; xj = xi + dx[0];
    draw_li(cr, gv->sensory_li,  xi, y[0], xj, y[0]);

    for (i = 0; i < NUM_MOTOR; i++) {
        for (j = i+1; j < NUM_MOTOR; j++) {
            double xi, xj;

            xi = x[1] + i * dx[1];
            xj = x[1] + j * dx[1];

            draw_li(cr, gv->response_li,  xi, y[1], xj, y[1]);
        }
    }

    /* The nodes: */
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    for (i = 0; i < NUM_SENSE; i++) {
        draw_node(cr, layout, x[0]+i*dx[0], y[0], gv, i, &(gv->sensory_node[i]));
    }
    for (i = 0; i < NUM_CONTEXT; i++) {
        draw_node(cr, layout, x[0]+(NUM_SENSE+i)*dx[0], y[0], gv, i, &(gv->context_node[i]));
    }
    for (i = 0; i < NUM_IMPERATIVE; i++) {
        draw_node(cr, layout, x[0]+(NUM_SENSE+NUM_CONTEXT+i)*dx[0], y[0], gv, i, &(gv->imperative_node[i]));
    }
    for (i = 0; i < NUM_MOTOR; i++) {
        draw_node(cr, layout, x[1]+i*dx[1], y[1], gv, i, &(gv->motor_node[i]));
    }

    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    for (i = 0; i < NUM_IMPERATIVE; i++) {
        if (gv->imperative_node[i].clamped) {
            CairoxPoint coordinates[2];

            coordinates[0].x = x[0]+(NUM_SENSE+NUM_CONTEXT+i)*dx[0];
            coordinates[0].y = y[0]-55;
            coordinates[1].x = x[0]+(NUM_SENSE+NUM_CONTEXT+i)*dx[0];
            coordinates[1].y = y[0]-20;
            cairox_paint_straight_arrow(cr, AH_SHARP, coordinates, 2, 2.0, 1.0);
        }
    }

    for (i = 0; i < NUM_CONTEXT; i++) {
        if (gv->context_node[i].clamped) {
            CairoxPoint coordinates[2];

            coordinates[0].x = x[0]+(NUM_SENSE+i)*dx[0];
            coordinates[0].y = y[0]-55;
            coordinates[1].x = x[0]+(NUM_SENSE+i)*dx[0];
            coordinates[1].y = y[0]-20;
            cairox_paint_straight_arrow(cr, AH_SHARP, coordinates, 2, 2.0, 1.0);
        }
    }
    
    for (i = 0; i < NUM_SENSE; i++) {
        if (gv->sensory_node[i].clamped) {
            CairoxPoint coordinates[2];

            coordinates[0].x = x[0]+i*dx[0];
            coordinates[0].y = y[0]-55;
            coordinates[1].x = x[0]+i*dx[0];
            coordinates[1].y = y[0]-20;
            cairox_paint_straight_arrow(cr, AH_SHARP, coordinates, 2, 2.0, 1.0);
        }
    }

    for (i = 0; i < NUM_MOTOR; i++) {
        if (gv->motor_node[i].clamped) {
            CairoxPoint coordinates[2];

            coordinates[0].x = x[1]+i*dx[1];
            coordinates[0].y = y[1]-55;
            coordinates[1].x = x[1]+i*dx[1];
            coordinates[1].y = y[1]-20;
            cairox_paint_straight_arrow(cr, AH_SHARP, coordinates, 2, 2.0, 1.0);
        }
    }

    i = 0;

    pangox_layout_set_font_size(layout, 14);
    cairox_text_parameters_set(&p, x[0]+(i+(NUM_SENSE-1)/2.0)*dx[0], y[0]-77, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Sensory Nodes");
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x[0]+i*dx[0], y[0]-30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Open"); i++;
    cairox_text_parameters_set(&p, x[0]+i*dx[0], y[0]-30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Close"); i++;

    pangox_layout_set_font_size(layout, 14);
    cairox_text_parameters_set(&p, x[0]+(i+(NUM_CONTEXT-1)/2.0)*dx[0], y[0]-77, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Context Nodes");
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x[0]+i*dx[0], y[0]-30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "1"); i++;
    cairox_text_parameters_set(&p, x[0]+i*dx[0], y[0]-30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "C1"); i++;
    cairox_text_parameters_set(&p, x[0]+i*dx[0], y[0]-30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "C2"); i++;
    cairox_text_parameters_set(&p, x[0]+i*dx[0], y[0]-30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "2"); i++;
    pangox_layout_set_font_size(layout, 14);
    cairox_text_parameters_set(&p, x[0]+(i+(NUM_IMPERATIVE-1)/2.0)*dx[0], y[0]-77, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Imperative Nodes");
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x[0]+i*dx[0], y[0]-30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "1"); i++;
    cairox_text_parameters_set(&p, x[0]+i*dx[0], y[0]-30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "2"); i++;

    pangox_layout_set_font_size(layout, 14);
    cairox_text_parameters_set(&p, width/2.0, y[1], PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Motor Nodes");
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x[1], y[1]+30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "OPEN");
    cairox_text_parameters_set(&p, x[1]+dx[1], y[1]+30, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "CLOSE");

//    pangox_layout_set_font_size(layout, 16);
//    cairox_text_parameters_set(&p, x[0]+1.0*dx[0], (y[0]+y[1])*0.5, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
//    cairox_paint_pango_text(cr, &p, layout, "Automatic\nRoute");
//    cairox_text_parameters_set(&p, x[0]+4.0*dx[0], (y[0]+y[1])*0.5, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_CENTER, 0.0);
//    cairox_paint_pango_text(cr, &p, layout, "Intentional\nRoute");


    if (gv->cycle > 0) {
        pangox_layout_set_font_size(layout, 14);
        g_snprintf(buffer, 64, "Cycle: %d", gv->cycle - TIME_ZERO);
        cairox_text_parameters_set(&p, width-5, 5, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_TOP, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

#ifdef DRAW_ASSOCIATION_TABLE
    /* Draw the table of association strengths: */
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
    cairox_paint_line(cr, 1.0, 10, height-50, 170-COL_WIDTH*0.5+(NUM_SENSE+NUM_CONTEXT+NUM_IMPERATIVE)*COL_WIDTH, height-50);
    cairox_paint_line(cr, 1.0, 10, height-36, 170-COL_WIDTH*0.5+(NUM_SENSE+NUM_CONTEXT+NUM_IMPERATIVE)*COL_WIDTH, height-36);
    cairox_paint_line(cr, 1.0, 10, height-3, 170-COL_WIDTH*0.5+(NUM_SENSE+NUM_CONTEXT+NUM_IMPERATIVE)*COL_WIDTH, height-3);
    cairox_paint_line(cr, 1.0, 10, height-51, 10, height-3);
    cairox_paint_line(cr, 1.0, 170-COL_WIDTH*0.5, height-51, 170-COL_WIDTH*0.5, height-3);
    cairox_paint_line(cr, 1.0, 170-COL_WIDTH*0.5+NUM_SENSE*COL_WIDTH, height-51, 170-COL_WIDTH*0.5+NUM_SENSE*COL_WIDTH, height-3);
    cairox_paint_line(cr, 1.0, 170-COL_WIDTH*0.5+(NUM_SENSE+NUM_CONTEXT)*COL_WIDTH, height-51, 170-COL_WIDTH*0.5+(NUM_SENSE+NUM_CONTEXT)*COL_WIDTH, height-3);
    cairox_paint_line(cr, 1.0, 170-COL_WIDTH*0.5+(NUM_SENSE+NUM_CONTEXT+NUM_IMPERATIVE)*COL_WIDTH, height-51, 170-COL_WIDTH*0.5+(NUM_SENSE+NUM_CONTEXT+NUM_IMPERATIVE)*COL_WIDTH, height-3);
    cairo_set_antialias(cr, CAIRO_ANTIALIAS_DEFAULT);

    pangox_layout_set_font_size(layout, 14);
    cairox_text_parameters_set(&p, 170+((NUM_SENSE-1)/2.0)*COL_WIDTH, height-34, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Sensory");
    cairox_text_parameters_set(&p, 170+(NUM_SENSE+(NUM_CONTEXT-1)/2.0)*COL_WIDTH, height-34, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Context");
    cairox_text_parameters_set(&p, 170+(NUM_SENSE+NUM_CONTEXT+(NUM_IMPERATIVE-1)/2.0)*COL_WIDTH, height-34, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Imperative");

    cairox_text_parameters_set(&p, 20, height-19, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "To OPEN / 1:");
    cairox_text_parameters_set(&p, 20, height-4, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "To CLOSE / 2:");

    for (j = 0; j < NUM_MOTOR; j++) {
        for (i = 0; i < NUM_SENSE; i++) {
            cairox_text_parameters_set(&p, 170+i*COL_WIDTH, height-19+j*15, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
            g_snprintf(buffer, 64, "%5.3f", WEIGHT_SM(j, i));
            cairox_paint_pango_text(cr, &p, layout, buffer);
        }

        for (i = 0; i < NUM_CONTEXT; i++) {
            cairox_text_parameters_set(&p, 170+(NUM_SENSE+i)*COL_WIDTH, height-19+j*15, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
            g_snprintf(buffer, 64, "%5.3f", WEIGHT_CM(j, i));
            cairox_paint_pango_text(cr, &p, layout, buffer);
        }

        for (i = 0; i < NUM_IMPERATIVE; i++) {
            cairox_text_parameters_set(&p, 170+(NUM_SENSE+NUM_CONTEXT+i)*COL_WIDTH, height-19+j*15, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
            g_snprintf(buffer, 64, "%5.3f", WEIGHT_IM(j, i));
            cairox_paint_pango_text(cr, &p, layout, buffer);
        }
    }
#endif

    g_object_unref(layout);
}

/******************************************************************************/
/* SECTION 2: Activation profile graphs ***************************************/
/******************************************************************************/

#define PMARGIN_LEFT   20
#define PMARGIN_RIGHT  10
#define PMARGIN_TOP     5
#define PMARGIN_BOTTOM 28

typedef enum {PROFILE_SENSORY, PROFILE_IMPERATIVE, PROFILE_MOTOR} ProfileGraphType;

static char *profile_graph_label(ProfileGraphType gt)
{
    if (gt == PROFILE_SENSORY) {
        return("Sensory Nodes");
    }
    else if (gt == PROFILE_IMPERATIVE) {
        return("Imperative Nodes");
    }
    else if (gt == PROFILE_MOTOR) {
        return("Motor Nodes");
    }
    else {
        return(NULL);
    }
}

static void draw_act_profile_graph(cairo_t *cr, PangoLayout *layout, GlobalVariables *gv, double x0, double y0, double x1, double y1, Node *node, int num_nodes, ProfileGraphType graph_type)
{
    CairoxTextParameters p;
    int i;
    char buffer[8];
    double x, y;
    int range;

    if (gv->cycle < 50) {
        range = 50;
    }
    else if (gv->cycle > MAX_CYCLES - 50) {
        range = MAX_CYCLES;
    }
    else {
        range = 50 * (1 + (gv->cycle / 50));
    }

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    pangox_layout_set_font_size(layout, 10);

    cairo_rectangle(cr, x0+PMARGIN_LEFT, y0+PMARGIN_TOP, (x1-x0-PMARGIN_LEFT-PMARGIN_RIGHT), (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM));
    cairo_stroke(cr);

     /* The response/habituation threshold: */
    if (graph_type == PROFILE_SENSORY) {
        cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
        y = y1-PMARGIN_BOTTOM - (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM) * gv->habituation_threshold;
        cairox_paint_line(cr, 1.0, x0+PMARGIN_LEFT, y, x1-PMARGIN_RIGHT, y);
    }
    else if (graph_type == PROFILE_MOTOR) {
        cairo_set_source_rgb(cr, 0.4, 0.4, 0.4);
        y = y1-PMARGIN_BOTTOM - (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM) * gv->actual_response_threshold;
        cairox_paint_line(cr, 1.0, x0+PMARGIN_LEFT, y, x1-PMARGIN_RIGHT, y);
    }
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    for (i = 0; i < 6; i++) {
        y = y1-PMARGIN_BOTTOM - (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM) * i / 5.0;
        cairox_paint_line(cr, 1.0, x0+PMARGIN_LEFT-3, y, x0+PMARGIN_LEFT+2, y);
        g_snprintf(buffer, 8, "%3.1f", (node[0].maths->max_act-node[0].maths->min_act) * i / 5.0);
        cairox_text_parameters_set(&p, x0+PMARGIN_LEFT-4, y, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    for (i = 0; i < 6; i++) {
        x = x0+PMARGIN_LEFT + (x1-x0-PMARGIN_LEFT-PMARGIN_RIGHT) * i / 5.0;
        cairox_paint_line(cr, 1.0, x, y1-PMARGIN_BOTTOM-2, x,  y1-PMARGIN_BOTTOM+3);
        g_snprintf(buffer, 8, "%3d", range * i / 5);
        cairox_text_parameters_set(&p, x, y1-PMARGIN_BOTTOM+4, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    /* Now draw each node's activation line: */

    for (i = 0; i < num_nodes; i++) {
        int c = 0;
        cairox_select_grey(cr, node[i].colour, 1.0);
        x = x0+PMARGIN_LEFT+(x1-x0-PMARGIN_LEFT-PMARGIN_RIGHT) * c / (double) range;
        y = y1-PMARGIN_BOTTOM - (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM) * node[i].activation[c] / (node[i].maths->max_act-node[i].maths->min_act);
        cairo_move_to(cr, x, y);
        for (c = 1; c < gv->cycle; c++) {
            x = x0+PMARGIN_LEFT+(x1-x0-PMARGIN_LEFT-PMARGIN_RIGHT) * c / (double) range;
            y = y1-PMARGIN_BOTTOM - (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM) * node[i].activation[c] / (node[i].maths->max_act-node[i].maths->min_act);
            cairo_line_to(cr, x, y);
        }
        cairo_stroke(cr);
    }

    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, (x0+x1)/2.0, y1-1, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, profile_graph_label(graph_type));
}

static void cairo_draw_act_profiles(cairo_t *cr, GlobalVariables *gv)
{
    PangoLayout *layout;
    int width, height;

    width = GTK_WIDGET(gv->canvas_act_profiles)->allocation.width;
    height = GTK_WIDGET(gv->canvas_act_profiles)->allocation.height;

    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    draw_act_profile_graph(cr, layout, gv, 0, 0.000 * height, width, 0.330 * height, gv->imperative_node, NUM_IMPERATIVE, PROFILE_IMPERATIVE);
    draw_act_profile_graph(cr, layout, gv, 0, 0.335 * height, width, 0.665 * height, gv->sensory_node, NUM_SENSE, PROFILE_SENSORY);
    draw_act_profile_graph(cr, layout, gv, 0, 0.670 * height, width, 1.000 * height, gv->motor_node, NUM_MOTOR, PROFILE_MOTOR);

    /* Discard the layout: */
    g_object_unref(layout);
}

/******************************************************************************/

static void draw_weight_profile_graph(cairo_t *cr, PangoLayout *layout, GlobalVariables *gv, double x0, double y0, double x1, double y1, int n)
{
    CairoxTextParameters p;
    int i, c;
    char buffer[64];
    double x, y;
    double range = 1.0;

    for (i = 0; i < (NUM_SENSE+NUM_CONTEXT); i++) {
        for (c = 0; c < gv->epochs; c++) {
            range = MAX(range, association[n][i].strength[c]);
        }
    }

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    pangox_layout_set_font_size(layout, 10);

    cairo_rectangle(cr, x0+PMARGIN_LEFT, y0+PMARGIN_TOP, (x1-x0-PMARGIN_LEFT-PMARGIN_RIGHT), (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM));
    cairo_stroke(cr);

    for (i = 0; i < 6; i++) {
        y = y1-PMARGIN_BOTTOM - (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM) * i / 5.0;
        cairox_paint_line(cr, 1.0, x0+PMARGIN_LEFT-3, y, x0+PMARGIN_LEFT+2, y);
        g_snprintf(buffer, 8, "%3.1f", i * 0.2 * range);
        cairox_text_parameters_set(&p, x0+PMARGIN_LEFT-4, y, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    for (i = 0; i < 7; i++) {
        x = x0+PMARGIN_LEFT + (x1-x0-PMARGIN_LEFT-PMARGIN_RIGHT) * i / 6.0;
        cairox_paint_line(cr, 1.0, x, y1-PMARGIN_BOTTOM-2, x,  y1-PMARGIN_BOTTOM+3);
        g_snprintf(buffer, 8, "%3d", gv->epochs * i / 6);
        cairox_text_parameters_set(&p, x, y1-PMARGIN_BOTTOM+4, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    /* Now draw each associations strength line: */

    for (i = 0; i < (NUM_SENSE+NUM_CONTEXT); i++) {
        c = 0;
        cairox_select_colour(cr, association[n][i].colour, 1.0);
        x = x0+PMARGIN_LEFT+(x1-x0-PMARGIN_LEFT-PMARGIN_RIGHT) * c / (double) (gv->epochs);
        y = y1-PMARGIN_BOTTOM - (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM) * association[n][i].strength[c] / range;
        cairo_move_to(cr, x, y);
        for (c = 1; c < gv->epochs; c++) {
            x = x0+PMARGIN_LEFT+(x1-x0-PMARGIN_LEFT-PMARGIN_RIGHT) * c / (double) (gv->epochs);
            y = y1-PMARGIN_BOTTOM - (y1-y0-PMARGIN_TOP-PMARGIN_BOTTOM) * association[n][i].strength[c] / range;
            cairo_line_to(cr, x, y);
        }
        cairo_stroke(cr);
    }

    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x1-PMARGIN_RIGHT-10, y0+PMARGIN_TOP+10, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_TOP, 0.0);
    g_snprintf(buffer, 64, "Assocations with response node %d", n+1);
    cairox_paint_pango_text(cr, &p, layout, buffer);
}

static void cairo_draw_weight_profiles(cairo_t *cr, GlobalVariables *gv)
{
    PangoLayout *layout;
    int width, height;

    width = GTK_WIDGET(gv->canvas_weight_profiles)->allocation.width;
    height = GTK_WIDGET(gv->canvas_weight_profiles)->allocation.height;

    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    draw_weight_profile_graph(cr, layout, gv, 0, 0.000 * height, width, 0.490 * height, 0);
    draw_weight_profile_graph(cr, layout, gv, 0, 0.500 * height, width, 0.990 * height, 1);

    /* Discard the layout: */
    g_object_unref(layout);
}

/******************************************************************************/

static void draw_error_bar(cairo_t *cr, double x0, double y0, double dy)
{
    if (dy > 0) {
        cairox_paint_line(cr, 1.0, x0, y0-dy, x0, y0+dy);
        cairox_paint_line(cr, 1.0, x0-3, y0-dy, x0+3, y0-dy);
        cairox_paint_line(cr, 1.0, x0-3, y0+dy, x0+3, y0+dy);
    }
}

static void draw_data_bar(cairo_t *cr, double y0, double height, double ymin, double ymax, double mean, double se, double x, int condition)
{
    cairo_rectangle(cr, x-10, y0, 20, -(mean-ymin) * height / (ymax-ymin));
    if (condition == 0) {
        cairo_set_source_rgb(cr, 0.9, 0.9, 0.9);
    }
    else {
        cairo_set_source_rgb(cr, 0.2, 0.2, 0.2);
    }
    cairo_fill_preserve(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_stroke(cr);
    draw_error_bar(cr, x, y0 - ((mean-ymin) * height / (ymax-ymin)),  (se * height / (ymax-ymin)));
}

static void draw_graph_exp1(cairo_t *cr, PangoLayout *layout, double x0, double x1, double y0, double y1, char *label, double mean[6], double se[6])
{
    CairoxTextParameters p;
    int ymin, ymax;
    int i;

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    cairo_rectangle(cr, x0+10, y0, x1-x0-10, y1-y0-10);
    cairo_stroke(cr);

    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, (x0+x1)/2.0, y1+8, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Training Group");
    pangox_layout_set_font_size(layout, 10);
    cairox_text_parameters_set(&p, x0+10+0.25*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Contingent");
    cairox_text_parameters_set(&p, x0+10+0.75*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Non-Contingent");

    ymin = mean[0]-se[0];
    for (i = 1; i < 4; i++) {
        ymin = MIN(ymin, mean[i]-se[i]);
    }
    ymin = (ymin / 50) * 50;
    ymax = ymin+120;

    pangox_layout_set_font_size(layout, 10);
    for (i = 0; i < 7; i++) {
        char buffer[16];
	double y = y0+i*(y1-y0-10)/6.0;

        g_snprintf(buffer, 16, "%d", ymin + (6-i) * (ymax-ymin)/6);
        cairox_text_parameters_set(&p, x0+8, y, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
        cairox_paint_line(cr, 1.0, x0+9, y, x0+13, y);
    }

    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[0], se[0], x0+10+0.20*(x1-x0-10), 0);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[1], se[1], x0+10+0.30*(x1-x0-10), 1);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[2], se[2], x0+10+0.70*(x1-x0-10), 0);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[3], se[3], x0+10+0.80*(x1-x0-10), 1);
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x0-28, (y0+y1)/2.0, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 90.0);
    cairox_paint_pango_text(cr, &p, layout, label);
}

static void draw_difference_graph_exp1(cairo_t *cr, PangoLayout *layout, double x0, double x1, double y0, double y1, char *label,  int ymin, int ymax, double mean[6], double se[6], int subjects)
{
    CairoxTextParameters p;
    int i;

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    cairo_rectangle(cr, x0+10, y0, x1-x0-10, y1-y0-10);
    cairo_stroke(cr);

    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, (x0+x1)/2.0, y1+8, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Training Group");
    pangox_layout_set_font_size(layout, 10);
    cairox_text_parameters_set(&p, x0+10+0.25*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Contingent");
    cairox_text_parameters_set(&p, x0+10+0.75*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Non-Contingent");

    pangox_layout_set_font_size(layout, 10);
    for (i = 0; i < 4; i++) {
        char buffer[16];
	double y = y0+i*(y1-y0-10)/3.0;

        g_snprintf(buffer, 16, "%d", ymin + (3-i) * (ymax-ymin)/3);
        cairox_text_parameters_set(&p, x0+8, y, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
        cairox_paint_line(cr, 1.0, x0+9, y, x0+13, y);
    }

    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[4], se[4], x0+10+0.25*(x1-x0-10), 0);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[5], se[5], x0+10+0.75*(x1-x0-10), 0);
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x0-28, (y0+y1)/2.0, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 90.0);
    cairox_paint_pango_text(cr, &p, layout, label);

    if (subjects > 1) {
        double diff = mean[5] - mean[4];
        double t, var_pooled;
        int df = 2 * (subjects-1);
        CairoxTextParameters p;
        char buffer[64];

        var_pooled = (se[4] * se[4] + se[5] * se[5]) * subjects * 0.5;
        t = (mean[5] - mean[4]) / sqrt(var_pooled * 2.0 / subjects);

        g_snprintf(buffer, 64, "Diff = %4.2f units; t(%d) = %6.3f; p = %6.4f", diff, df, t, prob_of_t(fabs(t), df));
        cairox_text_parameters_set(&p, (x0 + x1) * 0.5, y0+10, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }
}

static void draw_graph_exp2(cairo_t *cr, PangoLayout *layout, double x0, double x1, double y0, double y1, char *label, double mean[9], double se[9])
{
    CairoxTextParameters p;
    int ymin, ymax;
    int i;

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    cairo_rectangle(cr, x0+10, y0, x1-x0-10, y1-y0-10);
    cairo_stroke(cr);

    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, (x0+x1)/2.0, y1+8, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Training Group");
    pangox_layout_set_font_size(layout, 10);
    cairox_text_parameters_set(&p, x0+10+0.15*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Contingent");
    cairox_text_parameters_set(&p, x0+10+0.50*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Non-Contingent");
    cairox_text_parameters_set(&p, x0+10+0.85*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Signalled");

    ymin = mean[0]-se[0];
    for (i = 1; i < 6; i++) {
        ymin = MIN(ymin, mean[i]-se[i]);
    }
    ymin = (ymin / 50) * 50;
    ymax = ymin+120;

    // Hard coded graph range - comment this out for better flexibility
    ymin = 200;
    ymax = 350;

    pangox_layout_set_font_size(layout, 10);
    for (i = 0; i < 7; i++) {
        char buffer[16];
	double y = y0+i*(y1-y0-10)/6.0;

        g_snprintf(buffer, 16, "%d", ymin + (6-i) * (ymax-ymin)/6);
        cairox_text_parameters_set(&p, x0+8, y, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
        cairox_paint_line(cr, 1.0, x0+9, y, x0+13, y);
    }

    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[0], se[0], x0+10+0.11*(x1-x0-10), 0);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[1], se[1], x0+10+0.19*(x1-x0-10), 1);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[2], se[2], x0+10+0.46*(x1-x0-10), 0);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[3], se[3], x0+10+0.54*(x1-x0-10), 1);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[4], se[4], x0+10+0.81*(x1-x0-10), 0);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[5], se[5], x0+10+0.89*(x1-x0-10), 1);
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x0-28, (y0+y1)/2.0, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 90.0);
    cairox_paint_pango_text(cr, &p, layout, label);
}

static void draw_difference_graph_exp2(cairo_t *cr, PangoLayout *layout, double x0, double x1, double y0, double y1, char *label,  int ymin, int ymax, double mean[6], double se[6], int subjects)
{
    CairoxTextParameters p;
    int i;

    cairo_set_line_width(cr, 1.0);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    cairo_rectangle(cr, x0+10, y0, x1-x0-10, y1-y0-10);
    cairo_stroke(cr);

    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, (x0+x1)/2.0, y1+8, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Training Group");
    pangox_layout_set_font_size(layout, 10);
    cairox_text_parameters_set(&p, x0+10+0.15*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Contingent");
    cairox_text_parameters_set(&p, x0+10+0.50*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Non-Contingent");
    cairox_text_parameters_set(&p, x0+10+0.85*(x1-x0-10), y1-6, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Signalled");

    pangox_layout_set_font_size(layout, 10);
    for (i = 0; i < 5; i++) {
        char buffer[16];
	double y = y0+i*(y1-y0-10)/4.0;

        g_snprintf(buffer, 16, "%d", ymin + (4-i) * (ymax-ymin)/4);
        cairox_text_parameters_set(&p, x0+8, y, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
        cairox_paint_line(cr, 1.0, x0+9, y, x0+13, y);
    }

    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[6], se[6], x0+10+0.15*(x1-x0-10), 0);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[7], se[7], x0+10+0.50*(x1-x0-10), 0);
    draw_data_bar(cr, y1-10, y1-y0-10, ymin, ymax, mean[8], se[8], x0+10+0.85*(x1-x0-10), 0);
    pangox_layout_set_font_size(layout, 12);
    cairox_text_parameters_set(&p, x0-28, (y0+y1)/2.0, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 90.0);
    cairox_paint_pango_text(cr, &p, layout, label);

    if (subjects > 1) {
#if FALSE
        double diff = mean[5] - mean[4];
        double t, var_pooled;
        int df = 2 * (subjects-1);
        CairoxTextParameters p;
        char buffer[64];

        var_pooled = (se[4] * se[4] + se[5] * se[5]) * subjects * 0.5;
        t = (mean[5] - mean[4]) / sqrt(var_pooled * 2.0 / subjects);

        g_snprintf(buffer, 64, "Diff = %4.2f units; t(%d) = %6.3f; p = %6.4f", diff, df, t, prob_of_t(fabs(t), df));
        cairox_text_parameters_set(&p, (x0 + x1) * 0.5, y0+10, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
#endif
    }
}

static void draw_graph_legend(cairo_t *cr, PangoLayout *layout, double x0, double y0, int n)
{
    CairoxTextParameters p;
    char buffer[32];

    pangox_layout_set_font_size(layout, 10);
    g_snprintf(buffer, 32, "N = %d per condition", n);
    cairox_text_parameters_set(&p, x0+130, y0+5, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, buffer);

    cairo_rectangle(cr, x0, y0, 10, 10);
    cairo_set_source_rgb(cr, 0.9, 0.9, 0.9);
    cairo_fill_preserve(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_stroke(cr);

    cairo_rectangle(cr, x0, y0+18, 10, 10);
    cairo_set_source_rgb(cr, 0.2, 0.2, 0.2);
    cairo_fill_preserve(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_stroke(cr);

    pangox_layout_set_font_size(layout, 10);
    cairox_text_parameters_set(&p, x0+15, y0+5, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Compatible Trials");
    cairox_text_parameters_set(&p, x0+15, y0+23, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Incompatible Trials");
}

static void draw_graph_label(cairo_t *cr, PangoLayout *layout, double x0, double y0, GlobalVariables *gv)
{
    CairoxTextParameters p;
    char buffer[32];

    pangox_layout_set_font_size(layout, 10);
    g_snprintf(buffer, 32, "N = %d per condition", gv->subject_count2);
    cairox_text_parameters_set(&p, x0+130, y0+5, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, buffer);

    pangox_layout_set_font_size(layout, 10);
    g_snprintf(buffer, 32, "Learning rule: %s", lr_name[gv->learning_rule]);
    cairox_text_parameters_set(&p, x0, y0+5, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_CENTER, 0.0);
    cairox_paint_pango_text(cr, &p, layout, buffer);
}

static void exp1_cairo_draw_results(cairo_t *cr, GlobalVariables *gv)
{
    PangoLayout *layout;
    int width, height;
    double means[6], ses[6];

    width = GTK_WIDGET(gv->canvas1)->allocation.width;
    height = GTK_WIDGET(gv->canvas1)->allocation.height;

    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    /* RT: Actual Data */
    means[0] = 253.1; ses[0] = 12.3;
    means[1] = 270.6; ses[1] = 12.9;
    means[2] = 273.3; ses[2] = 15.2;
    means[3] = 298.9; ses[3] = 15.4;
    means[4] = 17.15; ses[4] = 2.9;
    means[5] = 25.30; ses[5] = 1.8;
    draw_graph_exp1(cr, layout, 0.05*width, 0.45*width, 0.05*height, 0.45*height, "RT (msec)", means, ses);
    draw_difference_graph_exp1(cr, layout, 0.55*width, 0.95*width, 0.05*height, 0.45*height, "Automatic Imitation (msec)", 0, 30, means, ses, 12);
    draw_graph_legend(cr, layout, 0.08*width, 0.06*height, 12);

    /* RT: Simulation Results */
    if (gv->subject_count1 > 0) {
        means[0] = exp1_get_condition_rt_mean(gv, TR_CONTINGENT, 0);
        ses[0] = exp1_get_condition_rt_se(gv, TR_CONTINGENT, 0);
        means[1] = exp1_get_condition_rt_mean(gv, TR_CONTINGENT, 1);
        ses[1] = exp1_get_condition_rt_se(gv, TR_CONTINGENT, 1);
        means[2] = exp1_get_condition_rt_mean(gv, TR_NONCONTINGENT, 0);
        ses[2] = exp1_get_condition_rt_se(gv, TR_NONCONTINGENT, 0);
        means[3] = exp1_get_condition_rt_mean(gv, TR_NONCONTINGENT, 1);
        ses[3] = exp1_get_condition_rt_se(gv, TR_NONCONTINGENT, 1);
        means[4] = exp1_get_condition_difference_mean(gv, TR_CONTINGENT);
        ses[4] = exp1_get_condition_difference_se(gv, TR_CONTINGENT);
        means[5] = exp1_get_condition_difference_mean(gv, TR_NONCONTINGENT);
        ses[5] = exp1_get_condition_difference_se(gv, TR_NONCONTINGENT);
        draw_graph_exp1(cr, layout, 0.05*width, 0.45*width, 0.55*height, 0.95*height, "RT (Cycles)", means, ses);
        draw_difference_graph_exp1(cr, layout, 0.55*width, 0.95*width, 0.55*height, 0.95*height, "Automatic Imitation (Cycles)", 0, 30, means, ses, gv->subject_count1);
        draw_graph_legend(cr, layout, 0.08*width, 0.56*height, gv->subject_count1);
    }

    /* Discard the layout: */
    g_object_unref(layout);
}

static void exp2_cairo_draw_results(cairo_t *cr, GlobalVariables *gv)
{
    PangoLayout *layout;
    int width, height;
    double means[9], ses[9];

    width = GTK_WIDGET(gv->canvas2)->allocation.width;
    height = GTK_WIDGET(gv->canvas2)->allocation.height;

    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    /* RT: Actual Data */
    means[0] = 252.4; ses[0] = 12.4;
    means[1] = 273.2; ses[1] = 13.8;
    means[2] = 250.6; ses[2] = 10.3;
    means[3] = 284.2; ses[3] = 13.3;
    means[4] = 235.7; ses[4] = 13.9;
    means[5] = 254.3; ses[5] = 13.7;
    means[6] = 20.8; ses[6] = 12.0 / sqrt(12.0);
    means[7] = 33.6; ses[7] = 16.9 / sqrt(12.0);
    means[8] = 18.6; ses[8] = 11.8 / sqrt(12.0);
    draw_graph_exp2(cr, layout, 0.05*width, 0.45*width, 0.05*height, 0.45*height, "RT (msec)", means, ses);
    draw_difference_graph_exp2(cr, layout, 0.55*width, 0.95*width, 0.05*height, 0.45*height, "Automatic Imitation (msec)", 0, 40, means, ses, 12);
    draw_graph_legend(cr, layout, 0.08*width, 0.06*height, 12);

    /* RT: Simulation Results */
    if (gv->subject_count2 > 0) {
        means[0] = exp2_get_condition_rt_mean(gv, TR_CONTINGENT, 0);
        ses[0] = exp2_get_condition_rt_se(gv, TR_CONTINGENT, 0);
        means[1] = exp2_get_condition_rt_mean(gv, TR_CONTINGENT, 1);
        ses[1] = exp2_get_condition_rt_se(gv, TR_CONTINGENT, 1);
        means[2] = exp2_get_condition_rt_mean(gv, TR_NONCONTINGENT, 0);
        ses[2] = exp2_get_condition_rt_se(gv, TR_NONCONTINGENT, 0);
        means[3] = exp2_get_condition_rt_mean(gv, TR_NONCONTINGENT, 1);
        ses[3] = exp2_get_condition_rt_se(gv, TR_NONCONTINGENT, 1);
        means[4] = exp2_get_condition_rt_mean(gv, TR_SIGNALLED, 0);
        ses[4] = exp2_get_condition_rt_se(gv, TR_SIGNALLED, 0);
        means[5] = exp2_get_condition_rt_mean(gv, TR_SIGNALLED, 1);
        ses[5] = exp2_get_condition_rt_se(gv, TR_SIGNALLED, 1);
        means[6] = exp2_get_condition_difference_mean(gv, TR_CONTINGENT);
        ses[6] = exp2_get_condition_difference_se(gv, TR_CONTINGENT);
        means[7] = exp2_get_condition_difference_mean(gv, TR_NONCONTINGENT);
        ses[7] = exp2_get_condition_difference_se(gv, TR_NONCONTINGENT);
        means[8] = exp2_get_condition_difference_mean(gv, TR_SIGNALLED);
        ses[8] = exp2_get_condition_difference_se(gv, TR_SIGNALLED);
        draw_graph_exp2(cr, layout, 0.05*width, 0.45*width, 0.55*height, 0.95*height, "RT (Cycles)", means, ses);
        draw_difference_graph_exp2(cr, layout, 0.55*width, 0.95*width, 0.55*height, 0.95*height, "Automatic Imitation (Cycles)", 0, 40, means, ses, gv->subject_count1);
        draw_graph_label(cr, layout, 0.08*width, 0.56*height, gv);
    }

    /* Discard the layout: */
    g_object_unref(layout);
}

/******************************************************************************/
/* SECTION 5: Update all visible canvases *************************************/
/******************************************************************************/

void update_graphics(GlobalVariables *gv)
{
    cairo_t *cr;

    if ((gv->canvas_network != NULL) && (GTK_WIDGET(gv->canvas_network)->window != NULL)) {
        cr = gdk_cairo_create(GTK_WIDGET(gv->canvas_network)->window);
        cairo_draw_network(cr, gv);
        cairo_destroy(cr);
    }

    if ((gv->canvas_act_profiles != NULL) && (GTK_WIDGET(gv->canvas_act_profiles)->window != NULL)) {
        cr = gdk_cairo_create(GTK_WIDGET(gv->canvas_act_profiles)->window);
        cairo_draw_act_profiles(cr, gv);
        cairo_destroy(cr);
    }

    if ((gv->canvas_weight_profiles != NULL) && (GTK_WIDGET(gv->canvas_weight_profiles)->window != NULL)) {
        cr = gdk_cairo_create(GTK_WIDGET(gv->canvas_weight_profiles)->window);
        cairo_draw_weight_profiles(cr, gv);
        cairo_destroy(cr);
    }

    if ((gv->canvas1 != NULL) && (GTK_WIDGET(gv->canvas1)->window != NULL)) {
        cr = gdk_cairo_create(GTK_WIDGET(gv->canvas1)->window);
        exp1_cairo_draw_results(cr, gv);
        cairo_destroy(cr);
    }

    if ((gv->canvas2 != NULL) && (GTK_WIDGET(gv->canvas2)->window != NULL)) {
        cr = gdk_cairo_create(GTK_WIDGET(gv->canvas2)->window);
        exp2_cairo_draw_results(cr, gv);
        cairo_destroy(cr);
    }

    gtkx_flush_events();
}

static void expose_canvas(GtkWidget *canvas, GdkEventExpose *event, GlobalVariables *gv)
{
    update_graphics(gv);
}

/******************************************************************************/

static void graphics_print_callback(GtkWidget *button, GlobalVariables *gv)
{
    cairo_surface_t *surface;
    int width, height;
    cairo_t *cr;

    if ((gv->canvas_network != NULL) && (GTK_WIDGET_MAPPED(gv->canvas_network))) {
        width = GTK_WIDGET(gv->canvas_network)->allocation.width;
        height = GTK_WIDGET(gv->canvas_network)->allocation.height;
        surface = cairo_pdf_surface_create("../Figures/image_network.pdf", width, height);
        cr = cairo_create(surface);
        cairo_draw_network(cr, gv);
        cairo_destroy(cr);
        cairo_surface_destroy(surface);
    }

    if ((gv->canvas_act_profiles != NULL) && (GTK_WIDGET_MAPPED(gv->canvas_act_profiles))) {
        width = GTK_WIDGET(gv->canvas_act_profiles)->allocation.width;
        height = GTK_WIDGET(gv->canvas_act_profiles)->allocation.height;
        surface = cairo_pdf_surface_create("../Figures/image_profiles_activations.pdf", width, height);
        cr = cairo_create(surface);
        cairo_draw_act_profiles(cr, gv);
        cairo_destroy(cr);
        cairo_surface_destroy(surface);
    }

    if ((gv->canvas_weight_profiles != NULL) && (GTK_WIDGET_MAPPED(gv->canvas_weight_profiles))) {
        width = GTK_WIDGET(gv->canvas_weight_profiles)->allocation.width;
        height = GTK_WIDGET(gv->canvas_weight_profiles)->allocation.height;
        surface = cairo_pdf_surface_create("../Figures/image_profiles_weights.pdf", width, height);
        cr = cairo_create(surface);
        cairo_draw_weight_profiles(cr, gv);
        cairo_destroy(cr);
        cairo_surface_destroy(surface);
    }

    if ((gv->canvas1 != NULL) && (GTK_WIDGET_MAPPED(gv->canvas1))) {
        width = GTK_WIDGET(gv->canvas1)->allocation.width;
        height = GTK_WIDGET(gv->canvas1)->allocation.height;
        surface = cairo_pdf_surface_create("../Figures/image_results_exp1.pdf", width, height);
        cr = cairo_create(surface);
        exp1_cairo_draw_results(cr, gv);
        cairo_destroy(cr);
        cairo_surface_destroy(surface);
    }

    if ((gv->canvas2 != NULL) && (GTK_WIDGET_MAPPED(gv->canvas2))) {
        width = GTK_WIDGET(gv->canvas2)->allocation.width;
        height = GTK_WIDGET(gv->canvas2)->allocation.height;
        surface = cairo_pdf_surface_create("../Figures/image_results_exp2.pdf", width, height);
        cr = cairo_create(surface);
        exp2_cairo_draw_results(cr, gv);
        cairo_destroy(cr);
        cairo_surface_destroy(surface);
    }
}

/******************************************************************************/
/* SECTION 6: */
/******************************************************************************/

#ifdef TRAINING
static void network_weights_train_callback(GtkWidget *button, GlobalVariables *gv)
{
    network_weights_train(gv, FALSE);
    update_graphics(gv);
}
#endif

static void network_weights_reset_callback(GtkWidget *button, GlobalVariables *gv)
{
    network_weights_initialise(gv);
    update_graphics(gv);
}

static void network_initialise_callback(GtkWidget *button, GlobalVariables *gv)
{
    gv->actual_response_threshold = gv->default_response_threshold;
    network_initialise_activations(gv, TRUE);
}

static void network_step_cycle_1_callback(GtkWidget *button, GlobalVariables *gv)
{
    network_step_cycle(gv, TRUE);
}

static void network_step_cycle_25_callback(GtkWidget *button, GlobalVariables *gv)
{
    int i;

    for (i = 0; i < 25; i++) {
        network_step_cycle(gv, TRUE);
    }
}

static void network_step_cycle_250_callback(GtkWidget *button, GlobalVariables *gv)
{
    int i;

    for (i = 0; i < 250; i++) {
        network_step_cycle(gv, TRUE);
    }
}

static void network_step_trial_callback(GtkWidget *button, GlobalVariables *gv)
{
    network_step_trial(gv, FALSE);
    update_graphics(gv);
}

/******************************************************************************/

static void set_spin_value_int(GtkWidget *button, int *value)
{
    *value = (int) gtk_spin_button_get_value(GTK_SPIN_BUTTON(button));
}

static void set_spin_value_double(GtkWidget *button, double *value)
{
    *value = (double) gtk_spin_button_get_value(GTK_SPIN_BUTTON(button));
}

static void create_spin_button_int(GtkWidget *box, char *label, int *value, int min, int max)
{
    GtkWidget *hbox, *tmp;

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new(label);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 5);
    gtk_widget_show(tmp);

    tmp = gtk_spin_button_new_with_range(min, max, 1.0);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(tmp), *value);
    g_signal_connect(G_OBJECT(tmp), "value-changed", G_CALLBACK(set_spin_value_int), value);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);
}

static void create_spin_button_double(GtkWidget *box, char *label, double *value, double min, double max, double step)
{
    GtkWidget *hbox, *tmp;

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new(label);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 5);
    gtk_widget_show(tmp);

    tmp = gtk_spin_button_new_with_range(min, max, step);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(tmp), *value);
    g_signal_connect(G_OBJECT(tmp), "value-changed", G_CALLBACK(set_spin_value_double), value);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);
}

static void set_check_button_value(GtkWidget *button, Boolean *value)
{
    *value = (Boolean) gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button));
}

static void create_checkbox(GtkWidget *box, char *label, Boolean *value)
{
    GtkWidget *hbox, *tmp;

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_check_button_new_with_label(label);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), *value);
    g_signal_connect(G_OBJECT(tmp), "toggled", G_CALLBACK(set_check_button_value), value);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);
}

/*----------------------------------------------------------------------------*/

#ifdef TRAINING

static void tr_set_contingent(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        network_set_training_data(gv, TR_CONTINGENT);
    }
}

static void tr_set_noncontingent(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        network_set_training_data(gv, TR_NONCONTINGENT);
    }
}

static void tr_set_signalled(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        network_set_training_data(gv, TR_SIGNALLED);
    }
}

static void tr_set_open_close(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        network_set_training_data(gv, TR_OPEN_CLOSE);
    }
}

static void tr_set_close_open(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        network_set_training_data(gv, TR_CLOSE_OPEN);
    }
}

static void tr_set_close(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        network_set_training_data(gv, TR_CLOSE);
    }
}

static void tr_set_open(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        network_set_training_data(gv, TR_OPEN);
    }
}

static void lr_set(GtkWidget *combobox, GlobalVariables *gv)
{
    gv->learning_rule = (LearningType) gtk_combo_box_get_active(GTK_COMBO_BOX(combobox));
}

#endif

static void clamp_s_off(GtkWidget *button, GlobalVariables *gv)
{
    gv->s_clamp = 0;
    update_graphics(gv);
}

static void clamp_s1(GtkWidget *button, GlobalVariables *gv)
{
    gv->s_clamp = 1;
    update_graphics(gv);
}

static void clamp_s2(GtkWidget *button, GlobalVariables *gv)
{
    gv->s_clamp = 2;
    update_graphics(gv);
}

static void clamp_i_off(GtkWidget *button, GlobalVariables *gv)
{
    gv->i_clamp = 0;
    update_graphics(gv);
}

static void clamp_i1(GtkWidget *button, GlobalVariables *gv)
{
    gv->i_clamp = 1;
    update_graphics(gv);
}

static void clamp_i2(GtkWidget *button, GlobalVariables *gv)
{
    gv->i_clamp = 2;
    update_graphics(gv);
}

/*----------------------------------------------------------------------------*/

static void network_run_exp1_callback(GtkWidget *button, GlobalVariables *gv)
{
    exp1_do_task(gv);
    exp1_print_analysis(stdout, gv);
}

/*----------------------------------------------------------------------------*/

static void network_run_exp2_callback(GtkWidget *button, GlobalVariables *gv)
{
    exp2_do_task(gv);
    exp2_print_analysis(stdout, gv);
}

/*----------------------------------------------------------------------------*/

static void rs_threshold(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        gv->selection = ST_THRESHOLD;
    }
}

static void rs_gap(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        gv->selection = ST_GAP;
    }
}

static void rs_probabilistic(GtkWidget *button, GlobalVariables *gv)
{
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(button))) {
        gv->selection = ST_PROBABILISTIC;
    }
}

GtkWidget *create_heading(GtkWidget *box, char *label)
{
    GtkWidget *hbox, *tmp;

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new(label);
    gtkx_widget_set_font_bold(tmp);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    return(hbox);
}

static void test_panel_create(GtkWidget *page, GlobalVariables *gv)
{
    GtkWidget *tmp, *hbox, *b, *panel;
    GSList *gs;

    panel = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(panel), 5);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(panel), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new(" Set imitative input:");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    b = gtk_radio_button_new_with_label(NULL, "off");
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(clamp_s_off), gv);
    gtk_widget_show(b);
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    b = gtk_radio_button_new_with_label(gs, "Open");
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(clamp_s1), gv);
    gtk_widget_show(b);
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    b = gtk_radio_button_new_with_label(gs, "Close");
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(clamp_s2), gv);
    gtk_widget_show(b);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(panel), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new(" Set imperative input:");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    b = gtk_radio_button_new_with_label(NULL, "off");
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(clamp_i_off), gv);
    gtk_widget_show(b);
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    b = gtk_radio_button_new_with_label(gs, "1 = Open");
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(clamp_i1), gv);
    gtk_widget_show(b);
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    b = gtk_radio_button_new_with_label(gs, "2 = Close");
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(clamp_i2), gv);
    gtk_widget_show(b);

    create_spin_button_int(panel, " SOA:", &(gv->soa), -200, 200);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(panel), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    gtkx_button_create(hbox, "Initialise", network_initialise_callback, gv);
    gtkx_button_create(hbox, "1 Step", network_step_cycle_1_callback, gv);
    gtkx_button_create(hbox, "25 Steps", network_step_cycle_25_callback, gv);
    gtkx_button_create(hbox, "250 Steps", network_step_cycle_250_callback, gv);
    gtkx_button_create(hbox, "Run", network_step_trial_callback, gv);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, TRUE, TRUE, 5);
    gtk_widget_show(tmp);

    gtkx_button_create(hbox, "Save Image", graphics_print_callback, gv);

    gtk_box_pack_end(GTK_BOX(page), panel, FALSE, FALSE, 0);
    gtk_widget_show(panel);
}

static void train_panel_create(GtkWidget *page, GlobalVariables *gv)
{
    GtkWidget *panel, *tmp, *hbox, *vleft, *vright, *b;
    GSList *gs;
    int i;

    panel = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(panel), 5);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(panel), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new("Training Set:");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    b = gtk_radio_button_new_with_label(NULL, "Contingent");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(tr_set_contingent), gv);
    gtk_widget_show(b);

    b = gtk_radio_button_new_with_label(gs, "NonContingent");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(tr_set_noncontingent), gv);
    gtk_widget_show(b);

    b = gtk_radio_button_new_with_label(gs, "Signalled");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(tr_set_signalled), gv);
    gtk_widget_show(b);

    b = gtk_radio_button_new_with_label(gs, "Open -> Close");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(tr_set_open_close), gv);
    gtk_widget_show(b);

    b = gtk_radio_button_new_with_label(gs, "Close -> Open");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(tr_set_close_open), gv);
    gtk_widget_show(b);

    b = gtk_radio_button_new_with_label(gs, "Open");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(tr_set_open), gv);
    gtk_widget_show(b);

    b = gtk_radio_button_new_with_label(gs, "Close");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(tr_set_close), gv);
    gtk_widget_show(b);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(panel), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    vleft = gtk_vbox_new(TRUE, 5);
    gtk_box_pack_start(GTK_BOX(hbox), vleft, TRUE, TRUE, 0);
    gtk_widget_show(vleft);

    vright = gtk_vbox_new(TRUE, 5);
    gtk_box_pack_start(GTK_BOX(hbox), vright, TRUE, TRUE, 0);
    gtk_widget_show(vright);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vleft), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new("Learning Rule:");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    tmp = gtk_combo_box_new_text();
    for (i = 0; i < NUM_LR; i++) {
        gtk_combo_box_append_text(GTK_COMBO_BOX(tmp), lr_name[i]);
    }
    gtk_combo_box_set_active(GTK_COMBO_BOX(tmp), gv->learning_rule);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(tmp), "changed", G_CALLBACK(lr_set), gv);
    gtk_widget_show(tmp);

    create_checkbox(vleft, "Print Deltas?", &gv->print_output);

    hbox = gtk_hbox_new(FALSE, 50);
    gtk_box_pack_start(GTK_BOX(vright), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    create_spin_button_double(hbox, "Weight Decay:", &(gv->weight_decay), 0.0, 1.0, 0.0001);
    create_spin_button_double(hbox, "Target/Limit:", &(gv->target_value), 0.0, 500.0, 0.1);

    hbox = gtk_hbox_new(FALSE, 50);
    gtk_box_pack_start(GTK_BOX(vright), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    create_spin_button_double(hbox, "Context Strength:", &(gv->context_input_strength[1]), 0.0, 100.0, 0.1);
    create_spin_button_double(hbox, "Learning Rate:", &(gv->learning_rate), 0.0, 1.0, 0.0001);

    hbox = gtk_hbox_new(FALSE, 140);
    gtk_box_pack_start(GTK_BOX(panel), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    create_spin_button_int(hbox, "Training Blocks:", &(gv->epochs), 1, TRAIN_MAX);
    gtkx_button_create(hbox, "Reset Weights", network_weights_reset_callback, gv);
    gtkx_button_create(hbox, "Train Weights", network_weights_train_callback, gv);

    gtk_box_pack_end(GTK_BOX(page), panel, FALSE, FALSE, 0);
    gtk_widget_show(panel);
}

static GtkWidget *net_train_panel_create(GlobalVariables *gv)
{
    GtkWidget *page;

    page = gtk_vbox_new(FALSE, 5);
    train_panel_create(page, gv);
    gtk_widget_show(page);
    return(page);
}

static GtkWidget *net_test_panel_create(GlobalVariables *gv)
{
    GtkWidget *page;

    page = gtk_vbox_new(FALSE, 5);
    test_panel_create(page, gv);
    gtk_widget_show(page);
    return(page);
}

static GtkWidget *network_canvas_create(GlobalVariables *gv)
{
    GtkWidget *page, *notebook;

    page = gtk_vbox_new(FALSE, 0);

    gv->canvas_network = gtk_drawing_area_new();
    g_signal_connect(G_OBJECT(gv->canvas_network), "expose_event", G_CALLBACK(expose_canvas), gv);
    gtk_box_pack_start(GTK_BOX(page), gv->canvas_network, TRUE, TRUE, 0);
    gtk_widget_show(gv->canvas_network);

    notebook = gtk_notebook_new();
    gtk_box_pack_start(GTK_BOX(page), notebook, FALSE, FALSE, 0);
    gtk_widget_show(notebook);

    gtkx_notebook_append_page(notebook, net_train_panel_create(gv), "Train");
    gtkx_notebook_append_page(notebook, net_test_panel_create(gv), "Test");

    return(page);
}

static GtkWidget *activation_profile_canvas_create(GlobalVariables *gv)
{
    GtkWidget *page, *tmp;

    page = gtk_vbox_new(FALSE, 5);

    gv->canvas_act_profiles = gtk_drawing_area_new();
    g_signal_connect(G_OBJECT(gv->canvas_act_profiles), "expose_event", G_CALLBACK(expose_canvas), gv);
    gtk_box_pack_start(GTK_BOX(page), gv->canvas_act_profiles, TRUE, TRUE, 0);
    gtk_widget_show(gv->canvas_act_profiles);

    /* And a separator with the controls below: */

    tmp = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(page), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    create_heading(page, " Test:");
    test_panel_create(page, gv);

    return(page);
}

static GtkWidget *weight_profile_canvas_create(GlobalVariables *gv)
{
    GtkWidget *page, *tmp;

    page = gtk_vbox_new(FALSE, 5);

    gv->canvas_weight_profiles = gtk_drawing_area_new();
    g_signal_connect(G_OBJECT(gv->canvas_weight_profiles), "expose_event", G_CALLBACK(expose_canvas), gv);
    gtk_box_pack_start(GTK_BOX(page), gv->canvas_weight_profiles, TRUE, TRUE, 0);
    gtk_widget_show(gv->canvas_weight_profiles);

    /* And a separator with the controls below: */

    tmp = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(page), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    create_heading(page, " Train:");
    train_panel_create(page, gv);

    return(page);
}

static GtkWidget *exp1_canvas_create(GlobalVariables *gv)
{
    GtkWidget *page, *tmp, *hbox;
    int i;

    page = gtk_vbox_new(FALSE, 0);

    gv->canvas1 = gtk_drawing_area_new();
    g_signal_connect(G_OBJECT(gv->canvas1), "expose_event", G_CALLBACK(expose_canvas), gv);
    gtk_box_pack_start(GTK_BOX(page), gv->canvas1, TRUE, TRUE, 0);
    gtk_widget_show(gv->canvas1);

    /* And a separator with the controls below: */

    tmp = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(page), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
    gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new("Learning Rule:");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    tmp = gtk_combo_box_new_text();
    for (i = 0; i < NUM_LR; i++) {
        gtk_combo_box_append_text(GTK_COMBO_BOX(tmp), lr_name[i]);
    }
    gtk_combo_box_set_active(GTK_COMBO_BOX(tmp), gv->learning_rule);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(tmp), "changed", G_CALLBACK(lr_set), gv);
    gtk_widget_show(tmp);

    gtkx_space(hbox);
    create_spin_button_double(hbox, "Context Strength:", &(gv->context_input_strength[1]), 0.0, 100.0, 0.1);
    gtkx_space(hbox);
    create_spin_button_double(hbox, "Learning Rate:", &(gv->learning_rate), 0.0, 1.0, 0.0001);
    gtkx_space(hbox);
//    create_spin_button_double(hbox, "Weight Decay:", &(gv->weight_decay), 0.0, 1.0, 0.0001);
//    gtkx_space(hbox);
    create_spin_button_double(hbox, "Target/Limit:", &(gv->target_value), 0.0, 500.0, 0.1);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
    gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    create_spin_button_int(hbox, "Group Size:", &(gv->subjects1), 1, MAX_SUBJECTS);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, TRUE, TRUE, 1);
    gtk_widget_show(tmp);

    create_spin_button_int(hbox, "Blocks/Subject:", &(gv->blocks1), 1, MAX_BLOCKS);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, TRUE, TRUE, 1);
    gtk_widget_show(tmp);

    gtkx_button_create(hbox, "Run", network_run_exp1_callback, gv);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, TRUE, TRUE, 1);
    gtk_widget_show(tmp);

    gtkx_button_create(hbox, "Save Image", graphics_print_callback, gv);

    return(page);
}

static GtkWidget *exp2_canvas_create(GlobalVariables *gv)
{
    GtkWidget *page, *tmp, *hbox;
    int i;

    page = gtk_vbox_new(FALSE, 0);

    gv->canvas2 = gtk_drawing_area_new();
    g_signal_connect(G_OBJECT(gv->canvas2), "expose_event", G_CALLBACK(expose_canvas), gv);
    gtk_box_pack_start(GTK_BOX(page), gv->canvas2, TRUE, TRUE, 0);
    gtk_widget_show(gv->canvas2);

    /* And a separator with the controls below: */

    tmp = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(page), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    hbox = gtk_hbox_new(FALSE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
    gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    tmp = gtk_label_new("Learning Rule:");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    tmp = gtk_combo_box_new_text();
    for (i = 0; i < NUM_LR; i++) {
        gtk_combo_box_append_text(GTK_COMBO_BOX(tmp), lr_name[i]);
    }
    gtk_combo_box_set_active(GTK_COMBO_BOX(tmp), gv->learning_rule);
    gtk_box_pack_start(GTK_BOX(hbox), tmp, FALSE, FALSE, 5);
    g_signal_connect(G_OBJECT(tmp), "changed", G_CALLBACK(lr_set), gv);
    gtk_widget_show(tmp);

    gtkx_space(hbox);
    create_spin_button_double(hbox, "Context Strength:", &(gv->context_input_strength[1]), 0.0, 100.0, 0.1);
    gtkx_space(hbox);
    create_spin_button_double(hbox, "Learning Rate:", &(gv->learning_rate), 0.0, 1.0, 0.0001);
    gtkx_space(hbox);
//    create_spin_button_double(hbox, "Weight Decay:", &(gv->weight_decay), 0.0, 1.0, 0.0001);
//    gtkx_space(hbox);
    create_spin_button_double(hbox, "Target/Limit:", &(gv->target_value), 0.0, 500.0, 0.1);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_set_border_width(GTK_CONTAINER(hbox), 5);
    gtk_box_pack_start(GTK_BOX(page), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    create_spin_button_int(hbox, "Group Size:", &(gv->subjects2), 1, MAX_SUBJECTS);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, TRUE, TRUE, 1);
    gtk_widget_show(tmp);

    create_spin_button_int(hbox, "Blocks/Subject:", &(gv->blocks2), 1, MAX_BLOCKS);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, TRUE, TRUE, 1);
    gtk_widget_show(tmp);

    gtkx_button_create(hbox, "Run", network_run_exp2_callback, gv);

    tmp = gtk_label_new(" ");
    gtk_box_pack_start(GTK_BOX(hbox), tmp, TRUE, TRUE, 1);
    gtk_widget_show(tmp);

    gtkx_button_create(hbox, "Save Image", graphics_print_callback, gv);

    return(page);
}

static GtkWidget *parameter_page_create(GlobalVariables *gv)
{
    GtkWidget *page, *hbox, *vbox, *tmp, *b;
    GSList *gs;

    page = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(page);

    /* Left side controls: */

    vbox = gtk_vbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(page), vbox, TRUE, TRUE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
    gtk_widget_show(vbox);

    create_heading(vbox, "Sensory Input Nodes:");
    create_spin_button_double(vbox, "Sensory Bias:", &gv->sensory_bias, -10.0, 10.0, 0.1);
    create_spin_button_double(vbox, "Sensory Input Strength:", &gv->sensory_input_strength, 0.0, 10.0, 0.1);
    create_spin_button_double(vbox, "Sensory Lateral Inhibition:", &gv->sensory_li, 0.0, 10.0, 0.1);
    create_spin_button_double(vbox, "Sensory Persistence:", &(gv->s_maths.persistence), 0.0, 1.0, 0.001);
    create_spin_button_double(vbox, "Sensory Noise (SD):", &(gv->s_maths.noise), 0.0, 10.0, 0.1);

    create_heading(vbox, "Imperative Input Nodes:");
    create_spin_button_double(vbox, "Imperative Bias:", &gv->imperative_bias, -10.0, 10.0, 0.1);
    create_spin_button_double(vbox, "Imperative Input Strength:", &gv->imperative_input_strength, 0.0, 10.0, 0.1);
    create_spin_button_double(vbox, "Imperative Persistence:", &(gv->d_maths.persistence), 0.0, 1.0, 0.001);
    create_spin_button_double(vbox, "Imperative Noise (SD):", &(gv->d_maths.noise), 0.0, 10.0, 0.1);

    create_heading(vbox, "Motor Response Nodes:");
    create_spin_button_double(vbox, "Motor Bias:", &gv->response_bias, -10.0, 10.0, 0.1);
    create_spin_button_double(vbox, "Motor Lateral Inhibition:", &gv->response_li, 0.0, 10.0, 0.1);
    create_spin_button_double(vbox, "Motor Persistence:", &(gv->r_maths.persistence), 0.0, 1.0, 0.001);
    create_spin_button_double(vbox, "Motor Noise (SD):", &(gv->r_maths.noise), 0.0, 10.0, 0.1);

    create_heading(vbox, "Time-course of Sensory Input:");
    create_spin_button_int(vbox, "Imitative Delay (in cycles):", &(gv->imitative_delay), -200, 200);
    create_checkbox(vbox, "Sensory Habituation?", &gv->sensory_habituation);
    create_checkbox(vbox, "Context Habituation?", &gv->context_habituation);
    create_spin_button_double(vbox, "Habituation Threshold:", &gv->habituation_threshold, 0.0, 1.0, 0.05);

    create_heading(vbox, "Response Selection:");

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);

    b = gtk_radio_button_new_with_label(NULL, "Threshold");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, TRUE, TRUE, 0);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(rs_threshold), gv);
    if (gv->selection == ST_THRESHOLD) {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(b), TRUE);
    }
    gtk_widget_show(b);
    b = gtk_radio_button_new_with_label(gs, "Gap");
    gs = gtk_radio_button_get_group(GTK_RADIO_BUTTON(b));
    gtk_box_pack_start(GTK_BOX(hbox), b, TRUE, TRUE, 0);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(rs_gap), gv);
    if (gv->selection == ST_GAP) {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(b), TRUE);
    }
    gtk_widget_show(b);
    b = gtk_radio_button_new_with_label(gs, "Probabilistic");
    gtk_box_pack_start(GTK_BOX(hbox), b, TRUE, TRUE, 0);
    g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(rs_probabilistic), gv);
    if (gv->selection == ST_PROBABILISTIC) {
        gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(b), TRUE);
    }
    gtk_widget_show(b);

    create_spin_button_double(vbox, "Selection Threshold/Gap:", &gv->default_response_threshold, 0.0, 1.0, 0.05);
    create_checkbox(vbox, "Dynamic Threshold Adjustment?", &gv->dynamic_response_threshold);

    /* Vertical separator: */

    tmp = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(page), tmp, FALSE, FALSE, 0);
    gtk_widget_show(tmp);

    /* Right side controls: */

    vbox = gtk_vbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(page), vbox, TRUE, TRUE, 0);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 5);
    gtk_widget_show(vbox);

    create_heading(vbox, "Weights:");
    create_spin_button_double(vbox, "Sensory -> Motor:", &gv->strength_sm, 0.0, 20.0, 0.1);
    create_spin_button_double(vbox, "Context -> Motor:", &gv->strength_cm, 0.0, 20.0, 0.1);
    create_spin_button_double(vbox, "Imperative -> Motor:", &gv->strength_im, 0.0, 20.0, 0.1);

    return(page);
}

/******************************************************************************/
/* SECTION N: Widget creation/initialisation routes ***************************/
/******************************************************************************/

static gint event_delete(GtkWidget *caller, GdkEvent *event, GlobalVariables *globals)
{
    /* Return 1 if we need to abort */
    return(0);
}

static void callback_destroy(GtkWidget *caller, GlobalVariables *gv)
{
    gtk_main_quit();
}

Boolean widgets_create(GlobalVariables *gv)
{
    if ((gv->window = gtk_window_new(GTK_WINDOW_TOPLEVEL)) != NULL) {
        GtkWidget *notebook;

        gtk_window_set_title(GTK_WINDOW(gv->window), "Mirror Learning");
        g_signal_connect(G_OBJECT(gv->window), "delete_event", G_CALLBACK(event_delete), gv);
        g_signal_connect(G_OBJECT(gv->window), "destroy", G_CALLBACK(callback_destroy), gv);
//        gtk_widget_set_size_request(gv->window, 550, -1);

        notebook = gtk_notebook_new();
        gtk_container_add(GTK_CONTAINER(gv->window), notebook);
        gtk_widget_show(notebook);

        gtkx_notebook_append_page(notebook, parameter_page_create(gv), "Network Parameters");
        gtkx_notebook_append_page(notebook, network_canvas_create(gv), "Network Structure");
        gtkx_notebook_append_page(notebook, activation_profile_canvas_create(gv), "Activation Profiles");
        gtkx_notebook_append_page(notebook, weight_profile_canvas_create(gv), "Weight Profiles");
        gtkx_notebook_append_page(notebook, exp1_canvas_create(gv), "Cook et al. Exp. 1");
        gtkx_notebook_append_page(notebook, exp2_canvas_create(gv), "Cook et al. Exp. 2");
   }
    return(gv->window != NULL);
}

/******************************************************************************/

static void graphs_draw(GlobalVariables *gv, cairo_t *cr, int width, int height)
{
    PangoLayout *layout;
    double means[9], ses[9];

    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    /* RT: Actual Data */
    means[0] = 252.4; ses[0] = 12.4;
    means[1] = 273.2; ses[1] = 13.8;
    means[2] = 250.6; ses[2] = 10.3;
    means[3] = 284.2; ses[3] = 13.3;
    means[4] = 235.7; ses[4] = 13.9;
    means[5] = 254.3; ses[5] = 13.7;
    means[6] = 20.8; ses[6] = 12.0 / sqrt(12.0);
    means[7] = 33.6; ses[7] = 16.9 / sqrt(12.0);
    means[8] = 18.6; ses[8] = 11.8 / sqrt(12.0);
    draw_graph_exp2(cr, layout, 0.05*width, 0.45*width, 0.02*height, 0.22*height, "RT (msec)", means, ses);
    draw_difference_graph_exp2(cr, layout, 0.55*width, 0.95*width, 0.02*height, 0.22*height, "Automatic Imitation (msec)", 0, 40, means, ses, 12);
    draw_graph_legend(cr, layout, 0.08*width, 0.03*height, 12);

    // We just want one ideal subject for these graphs:
    gv->subjects2 = 1; 
    gv->subject_im_sd = 0.0;

    /* RT: Simulation Results: Hebb */
    gv->learning_rule = LR_HEBB;
    gv->context_input_strength[2] = 60.0;
    gv->learning_rate = 0.0044;
    gv->target_value = 2.0;
    fprintf(stdout, "Running %s\n", lr_name[gv->learning_rule]);
    exp2_do_task(gv);
    means[0] = exp2_get_condition_rt_mean(gv, TR_CONTINGENT, 0);
    ses[0] = exp2_get_condition_rt_se(gv, TR_CONTINGENT, 0);
    means[1] = exp2_get_condition_rt_mean(gv, TR_CONTINGENT, 1);
    ses[1] = exp2_get_condition_rt_se(gv, TR_CONTINGENT, 1);
    means[2] = exp2_get_condition_rt_mean(gv, TR_NONCONTINGENT, 0);
    ses[2] = exp2_get_condition_rt_se(gv, TR_NONCONTINGENT, 0);
    means[3] = exp2_get_condition_rt_mean(gv, TR_NONCONTINGENT, 1);
    ses[3] = exp2_get_condition_rt_se(gv, TR_NONCONTINGENT, 1);
    means[4] = exp2_get_condition_rt_mean(gv, TR_SIGNALLED, 0);
    ses[4] = exp2_get_condition_rt_se(gv, TR_SIGNALLED, 0);
    means[5] = exp2_get_condition_rt_mean(gv, TR_SIGNALLED, 1);
    ses[5] = exp2_get_condition_rt_se(gv, TR_SIGNALLED, 1);
    means[6] = exp2_get_condition_difference_mean(gv, TR_CONTINGENT);
    ses[6] = exp2_get_condition_difference_se(gv, TR_CONTINGENT);
    means[7] = exp2_get_condition_difference_mean(gv, TR_NONCONTINGENT);
    ses[7] = exp2_get_condition_difference_se(gv, TR_NONCONTINGENT);
    means[8] = exp2_get_condition_difference_mean(gv, TR_SIGNALLED);
    ses[8] = exp2_get_condition_difference_se(gv, TR_SIGNALLED);
    draw_graph_exp2(cr, layout, 0.05*width, 0.45*width, 0.27*height, 0.47*height, "RT (Cycles)", means, ses);
    draw_difference_graph_exp2(cr, layout, 0.55*width, 0.95*width, 0.27*height, 0.47*height, "Automatic Imitation (Cycles)", 0, 40, means, ses, gv->subject_count1);
    draw_graph_label(cr, layout, 0.08*width, 0.28*height, gv);

    /* RT: Simulation Results: Qebb */
    gv->learning_rule = LR_QEBB;
    gv->context_input_strength[2] = 60.0;
    gv->learning_rate = 0.0054;
    gv->target_value = 20.0;
    fprintf(stdout, "Running %s\n", lr_name[gv->learning_rule]);
    exp2_do_task(gv);
    means[0] = exp2_get_condition_rt_mean(gv, TR_CONTINGENT, 0);
    ses[0] = exp2_get_condition_rt_se(gv, TR_CONTINGENT, 0);
    means[1] = exp2_get_condition_rt_mean(gv, TR_CONTINGENT, 1);
    ses[1] = exp2_get_condition_rt_se(gv, TR_CONTINGENT, 1);
    means[2] = exp2_get_condition_rt_mean(gv, TR_NONCONTINGENT, 0);
    ses[2] = exp2_get_condition_rt_se(gv, TR_NONCONTINGENT, 0);
    means[3] = exp2_get_condition_rt_mean(gv, TR_NONCONTINGENT, 1);
    ses[3] = exp2_get_condition_rt_se(gv, TR_NONCONTINGENT, 1);
    means[4] = exp2_get_condition_rt_mean(gv, TR_SIGNALLED, 0);
    ses[4] = exp2_get_condition_rt_se(gv, TR_SIGNALLED, 0);
    means[5] = exp2_get_condition_rt_mean(gv, TR_SIGNALLED, 1);
    ses[5] = exp2_get_condition_rt_se(gv, TR_SIGNALLED, 1);
    means[6] = exp2_get_condition_difference_mean(gv, TR_CONTINGENT);
    ses[6] = exp2_get_condition_difference_se(gv, TR_CONTINGENT);
    means[7] = exp2_get_condition_difference_mean(gv, TR_NONCONTINGENT);
    ses[7] = exp2_get_condition_difference_se(gv, TR_NONCONTINGENT);
    means[8] = exp2_get_condition_difference_mean(gv, TR_SIGNALLED);
    ses[8] = exp2_get_condition_difference_se(gv, TR_SIGNALLED);
    draw_graph_exp2(cr, layout, 0.05*width, 0.45*width, 0.52*height, 0.72*height, "RT (Cycles)", means, ses);
    draw_difference_graph_exp2(cr, layout, 0.55*width, 0.95*width, 0.52*height, 0.72*height, "Automatic Imitation (Cycles)", 0, 40, means, ses, gv->subject_count1);
    draw_graph_label(cr, layout, 0.08*width, 0.53*height, gv);

    /* RT: Simulation Results: RW */
    gv->learning_rule = LR_RW_A;
    gv->context_input_strength[2] = 30.0;
    gv->learning_rate = 0.0028;
    gv->target_value = 8.0;
    fprintf(stdout, "Running %s\n", lr_name[gv->learning_rule]);
    exp2_do_task(gv);
    means[0] = exp2_get_condition_rt_mean(gv, TR_CONTINGENT, 0);
    ses[0] = exp2_get_condition_rt_se(gv, TR_CONTINGENT, 0);
    means[1] = exp2_get_condition_rt_mean(gv, TR_CONTINGENT, 1);
    ses[1] = exp2_get_condition_rt_se(gv, TR_CONTINGENT, 1);
    means[2] = exp2_get_condition_rt_mean(gv, TR_NONCONTINGENT, 0);
    ses[2] = exp2_get_condition_rt_se(gv, TR_NONCONTINGENT, 0);
    means[3] = exp2_get_condition_rt_mean(gv, TR_NONCONTINGENT, 1);
    ses[3] = exp2_get_condition_rt_se(gv, TR_NONCONTINGENT, 1);
    means[4] = exp2_get_condition_rt_mean(gv, TR_SIGNALLED, 0);
    ses[4] = exp2_get_condition_rt_se(gv, TR_SIGNALLED, 0);
    means[5] = exp2_get_condition_rt_mean(gv, TR_SIGNALLED, 1);
    ses[5] = exp2_get_condition_rt_se(gv, TR_SIGNALLED, 1);
    means[6] = exp2_get_condition_difference_mean(gv, TR_CONTINGENT);
    ses[6] = exp2_get_condition_difference_se(gv, TR_CONTINGENT);
    means[7] = exp2_get_condition_difference_mean(gv, TR_NONCONTINGENT);
    ses[7] = exp2_get_condition_difference_se(gv, TR_NONCONTINGENT);
    means[8] = exp2_get_condition_difference_mean(gv, TR_SIGNALLED);
    ses[8] = exp2_get_condition_difference_se(gv, TR_SIGNALLED);
    draw_graph_exp2(cr, layout, 0.05*width, 0.45*width, 0.77*height, 0.97*height, "RT (Cycles)", means, ses);
    draw_difference_graph_exp2(cr, layout, 0.55*width, 0.95*width, 0.77*height, 0.97*height, "Automatic Imitation (Cycles)", 0, 40, means, ses, gv->subject_count1);
    draw_graph_label(cr, layout, 0.08*width, 0.78*height, gv);

    /* Discard the layout: */
    g_object_unref(layout);
}


void graphs_create(GlobalVariables *gv, int width, int height)
{
    cairo_surface_t *surface;
    cairo_t *cr;

    surface = cairo_pdf_surface_create("../Figures/image_graphs.pdf", width, height);
    cr = cairo_create(surface);
    graphs_draw(gv, cr, width, height);
    cairo_destroy(cr);
    cairo_surface_destroy(surface);
}

/******************************************************************************/
